# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'compoundscaleleft.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(335, 428)
        self.horizontalLayout = QtWidgets.QHBoxLayout(Form)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.command = Epc(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.command.sizePolicy().hasHeightForWidth())
        self.command.setSizePolicy(sizePolicy)
        self.command.setProperty("label_visible", True)
        self.command.setProperty("tx", True)
        self.command.setObjectName("command")
        self.verticalLayout.addWidget(self.command)
        self.echo = Epc(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.echo.sizePolicy().hasHeightForWidth())
        self.echo.setSizePolicy(sizePolicy)
        self.echo.setProperty("label_visible", True)
        self.echo.setProperty("tx", False)
        self.echo.setObjectName("echo")
        self.verticalLayout.addWidget(self.echo)
        self.numeric_status = Epc(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.numeric_status.sizePolicy().hasHeightForWidth())
        self.numeric_status.setSizePolicy(sizePolicy)
        self.numeric_status.setProperty("label_visible", True)
        self.numeric_status.setProperty("tx", False)
        self.numeric_status.setObjectName("numeric_status")
        self.verticalLayout.addWidget(self.numeric_status)
        self.horizontalLayout.addLayout(self.verticalLayout)
        self.status = Scale(Form)
        self.status.setObjectName("status")
        self.horizontalLayout.addWidget(self.status)
        self.horizontalLayout.setStretch(0, 2)
        self.horizontalLayout.setStretch(1, 3)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.command.setProperty("label_override", _translate("Form", "Command"))
        self.echo.setProperty("label_override", _translate("Form", "Response"))
        self.numeric_status.setProperty("label_override", _translate("Form", "Measured"))
from epyqlib.widgets.epc import Epc
from epyqlib.widgets.scale import Scale



#!/usr/bin/env python3

# TODO: get some docstrings in here!

# import can

# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


def referenced_files(raw_dict):
    return ()


class DeviceExtension:
    def __init__(self, device):
        self.device = device

    def post(self):
        pass

    def terminate(self):
        self.device = None


if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure



#!/usr/bin/env python3

#TODO: """DocString if there is one"""

import epyqlib.widgets.abstractwidget
import epyqlib.widgets.lcd_ui


# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class Lcd(epyqlib.widgets.abstractwidget.AbstractWidget):
    def __init__(self, parent=None, in_designer=False):
        super().__init__(
            ui_class=epyqlib.widgets.lcd_ui.Ui_Form,
            parent=parent,
            in_designer=in_designer,
        )

        self._frame = None
        self._signal = None

    def set_value(self, value):
        if self.signal_object is not None:
            if len(self.signal_object.enumeration) > 0:
                value = self.signal_object.full_string
            else:
                value = self.signal_object.format_float()
        elif value is None:
            # TODO: quit hardcoding this and it's better implemented elsewhere
            value = '{0:.2f}'.format(0)
        else:
            # TODO: quit hardcoding this and it's better implemented elsewhere
            value = '{0:.2f}'.format(value)

        self.ui.lcd.display(value)


if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure



#!/usr/bin/env python3

#TODO: """DocString if there is one"""

import epyqlib.widgets.abstractsignalpluginclass
import epyqlib.widgets.led

# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class LedPlugin(epyqlib.widgets.abstractsignalpluginclass.AbstractSignalPlugin):
    def __init__(self, parent=None):
        epyqlib.widgets.abstractsignalpluginclass.AbstractSignalPlugin.__init__(
            self, parent=parent)

        self._init = epyqlib.widgets.led.Led
        self._module_path = 'epyqlib.widgets.led'
        self._name = 'Led'


if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure



#!/usr/bin/env python3

#TODO: """DocString if there is one"""

# See file COPYING in this source tree
__copyright__ = ('\n'.join([
    'Copyright 2010, Kunal Anand',
    # https://gist.github.com/327585

    'Copyright 2011, Tennessee Carmel-Veilleux',
    # http://www.tentech.ca/2011/05/stream-tee-in-python-saving-stdout-to-file-while-keeping-the-console-alive/

    'Copyright 2016, EPC Power Corp.'
    # Adjusted to take a list of principals
]))
__license__ = 'GPLv2+'

class Tee:
    def __init__(self, principals):
        if len(principals) < 1:
            # TODO: be more specific with exception selection
            raise Exception('Tee requires at least one principal')

        self.principals = principals
        self.__missing_method_name = None # Hack!

    def __getattribute__(self, name):
        return object.__getattribute__(self, name)

    def __getattr__(self, name):
        self.__missing_method_name = name # Could also be a property
        return getattr(self, '__methodmissing__')

    def __methodmissing__(self, *args, **kwargs):
        call = getattr(self.principals[0], self.__missing_method_name, None)
        if call is not None:
            result = call(*args, **kwargs)
        else:
            result = None

        for principal in self.principals[1:]:
            call = getattr(principal, self.__missing_method_name, None)
            if call is not None:
                call(*args, **kwargs)

        return result


if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'variableselection.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(636, 473)
        Form.setProperty("update_from_can_file", False)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.logging_led = Led(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.logging_led.sizePolicy().hasHeightForWidth())
        self.logging_led.setSizePolicy(sizePolicy)
        self.logging_led.setProperty("label_from_enumeration", True)
        self.logging_led.setObjectName("logging_led")
        self.gridLayout.addWidget(self.logging_led, 0, 4, 1, 1)
        self.update_parameters_button = QtWidgets.QPushButton(Form)
        self.update_parameters_button.setObjectName("update_parameters_button")
        self.gridLayout.addWidget(self.update_parameters_button, 1, 2, 1, 1)
        self.save_selection_button = QtWidgets.QPushButton(Form)
        self.save_selection_button.setObjectName("save_selection_button")
        self.gridLayout.addWidget(self.save_selection_button, 1, 1, 1, 1)
        self.load_binary_button = QtWidgets.QPushButton(Form)
        self.load_binary_button.setObjectName("load_binary_button")
        self.gridLayout.addWidget(self.load_binary_button, 0, 0, 1, 1)
        self.reset_button = Button(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.reset_button.sizePolicy().hasHeightForWidth())
        self.reset_button.setSizePolicy(sizePolicy)
        self.reset_button.setProperty("label_visible", False)
        self.reset_button.setProperty("tx", True)
        self.reset_button.setObjectName("reset_button")
        self.gridLayout.addWidget(self.reset_button, 0, 2, 1, 1)
        self.load_selection_button = QtWidgets.QPushButton(Form)
        self.load_selection_button.setObjectName("load_selection_button")
        self.gridLayout.addWidget(self.load_selection_button, 0, 1, 1, 1)
        self.view = VariableSelectionView(Form)
        self.view.setObjectName("view")
        self.gridLayout.addWidget(self.view, 2, 0, 1, 5)
        self.configuration_is_valid_led = Led(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.configuration_is_valid_led.sizePolicy().hasHeightForWidth())
        self.configuration_is_valid_led.setSizePolicy(sizePolicy)
        self.configuration_is_valid_led.setProperty("label_from_enumeration", True)
        self.configuration_is_valid_led.setObjectName("configuration_is_valid_led")
        self.gridLayout.addWidget(self.configuration_is_valid_led, 1, 0, 1, 1)
        self.process_raw_log_button = QtWidgets.QPushButton(Form)
        self.process_raw_log_button.setObjectName("process_raw_log_button")
        self.gridLayout.addWidget(self.process_raw_log_button, 0, 3, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.logging_led.setProperty("label_override", _translate("Form", "Logging"))
        self.update_parameters_button.setText(_translate("Form", "Update Parameters"))
        self.save_selection_button.setText(_translate("Form", "Save Selection"))
        self.load_binary_button.setText(_translate("Form", "Load Binary"))
        self.reset_button.setProperty("label_override", _translate("Form", "Reset Log"))
        self.load_selection_button.setText(_translate("Form", "Load Selection"))
        self.configuration_is_valid_led.setProperty("label_override", _translate("Form", "Configuration Is Valid"))
        self.process_raw_log_button.setText(_translate("Form", "Process Raw Log"))
from epyqlib.variableselectionview import VariableSelectionView
from epyqlib.widgets.button import Button
from epyqlib.widgets.led import Led



#!/usr/bin/env python3

#TODO: """DocString if there is one"""

import io
import os

from PyQt5 import uic
from PyQt5.QtCore import pyqtProperty, QFile, QFileInfo, QTextStream
from PyQt5.QtWidgets import QWidget

# from epyqlib.compoundscale import CompoundScale
# from compoundscale import CompoundScale
import epyqlib.compoundscale


# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class CompoundScaleLeft(epyqlib.compoundscale.CompoundScale):
    def __init__(self, parent=None, in_designer=False):
        # QWidget.__init__(self, parent=parent)
        epyqlib.compoundscale.CompoundScale.__init__(self, parent, in_designer)

    def getPath(self):
        return os.path.join(QFileInfo.absolutePath(QFileInfo(__file__)),
                          'compoundscaleleft.ui')
if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure



# This file should be overwritten when a distribution build is done


try:
    from epyqlib._build_generated import *
except ImportError:
    build_system = None
    build_id = None
    build_number = None
    build_version = None
    job_id = None
    job_url = None



#!/usr/bin/env python3

# TODO: get some docstrings in here!

import contextlib
import logging
import time

import can
import can.interfaces.pcan
from epyqlib.canneo import QtCanListener
import epyqlib.utils.qt
from PyQt5 import QtCore
from PyQt5.QtWidgets import QApplication

# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class BusProxy:
    went_offline = epyqlib.utils.qt.Signal()

    def __init__(self, bus=None, timeout=0.1, transmit=True, filters=None,
                 auto_disconnect=True):
        self.filters = filters
        self.auto_disconnect = auto_disconnect

        self.timeout = timeout
        self.notifier = NotifierProxy(self)
        self.real_notifier = None
        self.tx_notifier = NotifierProxy(None)
        self.bus = None
        self.set_bus(bus)

        self._transmit = transmit

        self.flash_timer = QtCore.QTimer()
        self.flash_timer.setSingleShot(True)
        self.flash_timer.setInterval(10 * 1000)
        self.flash_timer.timeout.connect(self.stop_flashing)

    @contextlib.contextmanager
    def managed_real_bus(self, bustype, channel):
        real_bus = can.interface.Bus(bustype=bustype, channel=channel)
        try:
            self.set_bus(bus=real_bus)
            try:
                yield
            finally:
                self.terminate()
        finally:
            real_bus.shutdown()

    @property
    def transmit(self):
        return self._transmit

    @transmit.setter
    def transmit(self, transmit):
        self._transmit = transmit

    def send(self, msg, on_success=None):
        return self._send(msg, on_success=on_success)

    def send_passive(self, msg, on_success=None):
        return self._send(msg, on_success=on_success, passive=True)

    def _send(self, msg, on_success=None, passive=False):
        if self.bus is not None and (self._transmit or passive):
            # TODO: this (the silly sleep) is really hacky and shouldn't be needed but it seems
            #       to be to keep from forcing socketcan offbus.  the issue
            #       can be recreated with the following snippet.
            # import can
            # import time
            # bus = can.interface.Bus(bustype='socketcan', channel='can0')
            # msg = can.message.Message(arbitration_id=0x00FFAB80, bytearray([0, 0, 0, 0, 0, 0, 0, 0]))
            # for i in range(50):
            #   bus.send(msg)
            #   time.sleep(.0003)
            #
            #       which results in stuff like
            #
            # altendky@tp:/epc/bin$ can0; candump -L -x can0,#FFFFFFFF | grep -E '(0[04]FFAB(88|90|80)|can0 2)'
            # (1469135699.755374) can0 00FFAB80#0000000000000000
            # (1469135699.755462) can0 00FFAB80#0000000000000000
            # (1469135699.755535) can0 00FFAB80#0000000000000000
            # (1469135699.755798) can0 00FFAB80#0000000000000000
            # (1469135699.755958) can0 00FFAB80#0000000000000000
            # (1469135699.756132) can0 00FFAB80#0000000000000000
            # (1469135699.756446) can0 00FFAB80#0000000000000000
            # (1469135699.756589) can0 20000004#000C000000000000
            # (1469135699.756589) can0 20000004#0030000000000000
            # (1469135699.756731) can0 00FFAB80#0000000000000000
            # (1469135699.757004) can0 00FFAB80#0000000000000000
            # (1469135699.757187) can0 00FFAB80#0000000000000000
            # (1469135699.757308) can0 20000040#0000000000000000
            # (1469135699.757460) can0 00FFAB80#0000000000000000
            # (1469135699.757634) can0 00FFAB80#0000000000000000
            # (1469135699.757811) can0 00FFAB80#0000000000000000
            # (1469135699.757980) can0 00FFAB80#0000000000000000
            # (1469135699.758173) can0 00FFAB80#0000000000000000
            # (1469135699.758319) can0 00FFAB80#0000000000000000
            # (1469135699.758392) can0 00FFAB80#0000000000000000
            # (1469135699.758656) can0 00FFAB80#0000000000000000
            # (1469135699.758726) can0 00FFAB80#0000000000000000
            # (1469135699.758894) can0 00FFAB80#0000000000000000

            if isinstance(self.bus, can.BusABC):
                # TODO: this is a hack to allow detection of transmitted
                #       messages later
                msg.timestamp = None

                try:
                    # TODO: I would use message=message (or msg=msg) but:
                    #       https://bitbucket.org/hardbyte/python-can/issues/52/inconsistent-send-signatures
                    self.bus.send(msg)
                except can.CanError:
                    # TODO: specifically implemented for a transmit queue
                    #       full situation to avoid infinite dialogs
                    self.set_bus()

                # TODO: get a real value for sent, but for now python-can
                #       doesn't provide that info.  also, it would be async...
                sent = True

                time.sleep(0.0005)

                if sent:
                    self.tx_notifier.message_received(message=msg)

                    if on_success is not None:
                        on_success()
            else:
                # TODO: I would use message=message (or msg=msg) but:
                #       https://bitbucket.org/hardbyte/python-can/issues/52/inconsistent-send-signatures
                sent = self.bus._send(msg, on_success=on_success, passive=passive)

            if self.auto_disconnect:
                self.verify_bus_ok()

            # TODO: since send() doesn't always report failures this won't either
            #       fix that
            return sent

        return False

    def verify_bus_ok(self):
        if self.bus is None:
            # No bus, nothing to go wrong with it... ?
            ok = True
        else:
            if hasattr(self.bus, 'StatusOk'):
                ok = self.bus.StatusOk()

                if not ok:
                    self.set_bus()
            else:
                ok = True
                # TODO: support socketcan
                if hasattr(self.bus, 'verify_bus_ok'):
                    ok = self.bus.verify_bus_ok()

        return ok

    def shutdown(self):
        pass

    def flash(self):
        if self.bus is not None:
            result = self.bus.flash(True)
            self.flash_timer.start()
            return result

    def stop_flashing(self):
        if self.bus is not None:
            return self.bus.flash(False)

    def terminate(self):
        self.set_bus()
        logging.debug('{} terminated'.format(object.__repr__(self)))

    def set_bus(self, bus=None):
        was_online = self.bus is not None

        if was_online:
            if isinstance(self.bus, can.BusABC):
                self.real_notifier.stop()
                time.sleep(1.1 * self.timeout)
            else:
                self.bus.notifier.remove(self.notifier)
                self.bus.tx_notifier.remove(self.tx_notifier)
            self.bus.shutdown()
        self.bus = bus

        if self.bus is not None:
            self.set_filters(self.filters)
            if isinstance(self.bus, can.BusABC):
                self.real_notifier = can.Notifier(
                    bus=self.bus,
                    listeners=[self.notifier],
                    timeout=self.timeout)
            else:
                self.bus.notifier.add(self.notifier)
                self.bus.tx_notifier.add(self.tx_notifier)
                self.real_notifier = None
        else:
            self.real_notifier = None

        self.reset()

        if was_online and self.bus is None:
            self.went_offline.emit()

        if isinstance(self.bus, can.BusABC):
            self.notifier.move_to_thread(None)
            self.tx_notifier.move_to_thread(None)
        else:
            app = QApplication.instance()
            if app is not None:
                self.notifier.move_to_thread(app.thread())
                self.tx_notifier.move_to_thread(app.thread())

    def reset(self):
        if self.bus is not None:
            if isinstance(self.bus, can.interfaces.pcan.PcanBus):
                self.bus.reset()
                # TODO: do this a better way
                # Give PCAN a chance to actually reset and avoid immediate
                # send failures
                time.sleep(0.500)
            else:
                # TODO: support socketcan
                if hasattr(self.bus, 'reset'):
                    self.bus.reset()

    def set_filters(self, filters):
        self.filters = filters
        real_bus = self.bus
        if real_bus is not None:
            if hasattr(real_bus, 'setFilters'):
                real_bus.setFilters(can_filters=filters)


class NotifierProxy(QtCanListener):
    def __init__(self, bus, listeners=[], filtered_ids=None, parent=None):
        super().__init__(receiver=self.message_received, parent=parent)

        # TODO: consider a WeakSet, though this may presently
        #       be keeping objects alive
        self.listeners = set(listeners)
        if filtered_ids is None:
            self.filtered_ids = None
        else:
            self.filtered_ids = set(filtered_ids)

    def message_received(self, message):
        if (self.filtered_ids is None or
                message.arbitration_id in self.filtered_ids):
            for listener in tuple(self.listeners):
                listener.message_received_signal.emit(message)

    def add(self, listener):
        self.listeners.add(listener)

    def discard(self, listener):
        self.listeners.discard(listener)

    def remove(self, listener):
        self.listeners.remove(listener)


if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure



#!/usr/bin/env python3

#TODO: """DocString if there is one"""

from PyQt5.QtCore import pyqtProperty
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QWidget

import epyqlib.compoundtoggle_ui
import epyqlib.utils.qt


# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class CompoundToggle(QWidget):
    def __init__(self, parent=None, in_designer=False):
        QWidget.__init__(self, parent=parent)

        self.in_designer = in_designer

        self.ui = epyqlib.compoundtoggle_ui.Ui_Form()
        self.ui.setupUi(self)

        self.ui.command.in_designer = self.in_designer
        self.ui.status_on.in_designer = self.in_designer
        self.ui.status_off.in_designer = self.in_designer

        if self.in_designer:
            self.ui.status_off.set_value(self.ui.status_off.on_value)
            self.ui.status_on.set_value(self.ui.status_off.on_value)

    @pyqtProperty(str)
    def box_title(self):
        return self.ui.box.title()

    @box_title.setter
    def box_title(self, title):
        self.ui.box.setTitle(title)

    @pyqtProperty('QString')
    def command_signal_path(self):
        return self.ui.command.signal_path

    @command_signal_path.setter
    def command_signal_path(self, value):
        self.ui.command.signal_path = value

    @pyqtProperty('QString')
    def status_signal_path(self):
        return self.ui.status_off.signal_path

    @status_signal_path.setter
    def status_signal_path(self, value):
        self.ui.status_off.signal_path = value
        self.ui.status_on.signal_path = value

    @pyqtProperty(int)
    def on_value(self):
        return self.ui.status_on.on_value

    @on_value.setter
    def on_value(self, new_on_value):
        self.ui.status_on.on_value = new_on_value

    @pyqtProperty(bool)
    def status_on_automatic_off_color(self):
        return self.ui.status_on.automatic_off_color

    @status_on_automatic_off_color.setter
    def status_on_automatic_off_color(self, automatic):
        self.ui.status_on.automatic_off_color = automatic

    @pyqtProperty(QColor)
    def status_on_on_color(self):
        return self.ui.status_on.on_color

    @status_on_on_color.setter
    def status_on_on_color(self, new_color):
        self.ui.status_on.on_color = new_color

    @pyqtProperty(QColor)
    def status_on_manual_off_color(self):
        return self.ui.status_on.manual_off_color

    @status_on_manual_off_color.setter
    def status_on_manual_off_color(self, new_color):
        self.ui.status_on.manual_off_color = new_color

    @pyqtProperty(int)
    def off_value(self):
        return self.ui.status_off.on_value

    @off_value.setter
    def off_value(self, new_off_value):
        self.ui.status_off.on_value = new_off_value
        if self.in_designer:
            self.ui.status_off.set_value(new_off_value)
            self.ui.status_on.set_value(new_off_value)

    @pyqtProperty(bool)
    def status_off_automatic_off_color(self):
        return self.ui.status_off.automatic_off_color

    @status_off_automatic_off_color.setter
    def status_off_automatic_off_color(self, automatic):
        self.ui.status_off.automatic_off_color = automatic

    @pyqtProperty(QColor)
    def status_off_on_color(self):
        return self.ui.status_off.on_color

    @status_off_on_color.setter
    def status_off_on_color(self, new_color):
        self.ui.status_off.on_color = new_color

    @pyqtProperty(QColor)
    def status_off_manual_off_color(self):
        return self.ui.status_off.manual_off_color
    
    @status_off_manual_off_color.setter
    def status_off_manual_off_color(self, new_color):
        self.ui.status_off.manual_off_color = new_color


if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure



import weakref

import attr
from PyQt5.QtCore import pyqtSlot, Qt, QCoreApplication, QEvent, QPoint
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtGui import QMouseEvent

import epyqlib.utils.qt


class Dispatch(QtWidgets.QStyledItemDelegate):
    def __init__(self, selector, parent):
        super().__init__(parent)

        self.selector = selector
        self.delegates = set()

        self.delegates_by_editor = weakref.WeakKeyDictionary()

    def createEditor(self, parent, option, index):
        delegate = self.selector(index)

        if delegate not in self.delegates:
            self.delegates.add(delegate)
            delegate.closeEditor.connect(self.closeEditor)
            delegate.commitData.connect(self.commitData)

        editor = delegate.createEditor(parent, option, index)

        self.delegates_by_editor[editor] = delegate

        return editor

    def destroyEditor(self, editor, *args, **kwargs):
        delegate = self.delegates_by_editor[editor]
        return delegate.destroyEditor(editor, *args, **kwargs)

    def setEditorData(self, editor, *args, **kwargs):
        delegate = self.delegates_by_editor[editor]
        return delegate.setEditorData(editor, *args, **kwargs)

    def setModelData(self, editor, *args, **kwargs):
        delegate = self.delegates_by_editor[editor]
        return delegate.setModelData(editor, *args, **kwargs)

    def updateEditorGeometry(self, editor, *args, **kwargs):
        delegate = self.delegates_by_editor[editor]
        return delegate.updateEditorGeometry(editor, *args, **kwargs)


def default(model, node, column):
    if hasattr(node, 'enumeration_strings'):
        if len(node.enumeration_strings()) > 0:
            return Delegate(creator=create_combo)

    if hasattr(node, 'secret'):
        if node.secret:
            return Delegate(modifier=modify_password)

    return Delegate()


@attr.s
class Delegate:
    creator = attr.ib(default=None)
    modifier = attr.ib(default=None)
    editor_setter = attr.ib(default=None)
    model_setter = attr.ib(default=None)


# TODO: CAMPid 374895478431714307074310
class CustomCombo(QtWidgets.QComboBox):
    def hidePopup(self):
        super().hidePopup()

        QtCore.QCoreApplication.postEvent(
            self,
            QtGui.QKeyEvent(
                QtCore.QEvent.KeyPress,
                QtCore.Qt.Key_Enter,
                QtCore.Qt.NoModifier,
            ),
        )


class ByFunction(QtWidgets.QStyledItemDelegate):
    def __init__(self, model, parent, function=default):
        QtWidgets.QStyledItemDelegate.__init__(self, parent=parent)

        self.model = model
        self.function = function

    def get_delegate_node(self, index):
        index = epyqlib.utils.qt.resolve_index_to_model(index)
        # TODO: way too particular
        node = self.model.node_from_index(index)

        delegate = self.function(
            model=self.model,
            node=node,
            column=index.column(),
        )

        return delegate, node

    def createEditor(self, parent, option, index):
        delegate, node = self.get_delegate_node(index=index)

        if delegate.creator is None:
            widget = super().createEditor(parent, option, index)
        else:
            widget = delegate.creator(index=index, node=node, parent=parent)

        if delegate.modifier is not None:
            delegate.modifier(widget=widget)

        return widget

    # def setEditorData(self, editor, index):
    #     delegate, node = self.get_delegate_node(index=index)
    #
    #     if delegate.editor_setter is None:
    #         return super().setEditorData(editor, index)
    #
    #     return delegate.editor_setter(editor=editor, index=index)

    def setModelData(self, editor, model, index):
        delegate, node = self.get_delegate_node(index=index)

        if delegate.model_setter is None:
            return super().setModelData(editor, model, index)

        return delegate.model_setter(
            editor=editor,
            model=model,
            index=index,
        )


def create_combo(index, node, parent):
    widget = CustomCombo(parent=parent)

    # TODO: use the userdata to make it easier to get in and out
    widget.addItems(node.enumeration_strings(include_values=True))

    present_string = str(node.fields[index.column()])
    index = widget.findText(present_string)
    if index == -1:
        widget.setCurrentIndex(0)
    else:
        widget.setCurrentIndex(index)

    view = widget.view()
    view.setMinimumWidth(calculate_combo_view_width(widget))

    event = QMouseEvent(QEvent.MouseButtonPress,
                        QPoint(),
                        Qt.LeftButton,
                        Qt.LeftButton,
                        Qt.NoModifier)
    QCoreApplication.postEvent(widget, event)

    return widget


def calculate_combo_view_width(widget):
    view = widget.view()
    metric = view.fontMetrics()
    scrollbar = view.verticalScrollBar()

    scrollbar_width = 0
    if scrollbar.isVisibleTo(view):
        scrollbar_width = scrollbar.width()

    text = (widget.itemText(i) for i in range(widget.count()))
    text_width = max(metric.width(s) for s in text)

    # consider width of icons, for example

    return text_width + scrollbar_width


def create_button(index, node, parent):
    text = node.button_text(index.column())
    widget = QtWidgets.QPushButton(parent=parent)
    widget.setText(text)

    return widget


def modify_password(widget):
    widget.setEchoMode(QtWidgets.QLineEdit.Password)



import pathlib

import attr
import PyQt5.uic


Ui, UiBase = PyQt5.uic.loadUiType(
    pathlib.Path(__file__).with_suffix('.ui'),
)


# TODO: what about `in_designer=False`


@attr.s
class FilesView(UiBase):
    ui = attr.ib(factory=Ui)

    @classmethod
    def build(cls):
        instance = cls()
        instance.setup_ui()

        return instance

    @classmethod
    def qt_build(cls, parent, in_designer=False):
        instance = cls.build()
        instance.setParent(parent)

        return instance

    def __attrs_post_init__(self):
        super().__init__()

    def setup_ui(self):
        self.ui.setupUi(self)


# .ui files need a direct module attribute, not a class method, afaict.
FilesViewQtBuilder = FilesView.qt_build



#!/usr/bin/env python3

import os
import subprocess

try:
    hash = subprocess.check_output([os.path.join('c:/', 'Program Files', 'Git', 'bin', 'git.exe'), 'rev-parse', 'HEAD'])
except FileNotFoundError:
    hash = subprocess.check_output(['git', 'rev-parse', 'HEAD'])

out_file = 'revision.py'

with open(out_file, 'w') as file:
    file.write("hash = '{}'\n".format(hash.decode('utf-8').rstrip()))



#!/usr/bin/env python3

#TODO: """DocString if there is one"""

import math
import textwrap

import can
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QHeaderView, QMenu, QMessageBox
from PyQt5.QtCore import Qt, pyqtSignal, QItemSelectionModel

import epyqlib.datalogger
import epyqlib.device
import epyqlib.devicetree
import epyqlib.devicetreeview_ui
import epyqlib.flash
import epyqlib.utils.general
import epyqlib.utils.qt
import epyqlib.utils.twisted

# See file COPYING in this source tree
__copyright__ = 'Copyright 2017, EPC Power Corp.'
__license__ = 'GPLv2+'


class UnsupportedError(Exception):
    pass


def load_device(bus=None, file=None, parent=None):
    # TODO  CAMPid 9561616237195401795426778
    if file is None:
        filters = [
            ('EPC Packages', ['epc', 'epz']),
            ('All Files', ['*'])
        ]
        file = epyqlib.utils.qt.file_dialog(filters, parent=parent)

        if file is None:
            return

    try:
        return epyqlib.device.Device(file=file, bus=bus)
    except epyqlib.device.CancelError:
        pass

    return


class DeviceTreeView(QtWidgets.QWidget):
    device_selected = pyqtSignal(epyqlib.device.Device)

    def __init__(self, parent=None, in_designer=False):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        self.ui = epyqlib.devicetreeview_ui.Ui_Form()
        self.ui.setupUi(self)

        self.resize_columns = epyqlib.devicetree.Columns(
            name=True,
            nickname=True,
            bitrate=True,
            transmit=True,
        )

        self.ui.tree_view.setContextMenuPolicy(Qt.CustomContextMenu)
        self.ui.tree_view.customContextMenuRequested.connect(
            self.context_menu
        )

        self.model = None

    def _current_changed(self, new_index, old_index):
        node = self.model.node_from_index(new_index)
        if isinstance(node, epyqlib.devicetree.Device):
            device = node.device
            self.device_selected.emit(device)

    def context_menu(self, position):
        index = self.ui.tree_view.indexAt(position)

        if not index.isValid():
            return

        node = self.model.node_from_index(index)

        add_device_action = None
        remove_device_action = None
        flash_action = None
        pull_raw_log_action = None
        check_compatibility_action = None
        write_to_epz_action = None
        blink_action = None

        menu = QMenu()
        if isinstance(node, epyqlib.devicetree.Device):
            offline = (not node.tree_parent._checked.name
                      and not node.tree_parent.fields.name == 'Offline')

            pull_raw_log_action = menu.addAction('Pull Raw Log...')
            pull_raw_log_action.setEnabled(offline)

            write_to_epz_action = menu.addAction('Write To Epz...')
            write_to_epz_action.setEnabled(not node.device.from_zip)
            # TODO: stdlib zipfile can't create an encrypted .zip
            #       make a good solution that will...
            write_to_epz_action.setVisible(False)

            check_compatibility_action = menu.addAction('Check Compatibility')
            check_compatibility_action.setEnabled(not offline)

            remove_device_action = menu.addAction('Close')
        if isinstance(node, epyqlib.devicetree.Bus):
            add_device_action = menu.addAction('Open device file...')
            flash_action = menu.addAction('Load firmware...')
            flash_action.setEnabled(
                not node._checked.name
                and not node.fields.name == 'Offline'
            )
            blink_action = menu.addAction('Blink adapter')
            blink_action.setEnabled(
                node.interface == 'pcan'
                and node.bus.bus is not None
            )

        action = menu.exec_(self.ui.tree_view.viewport().mapToGlobal(position))

        if action is None:
            pass
        elif action is remove_device_action:
            self.remove_device(node)
        elif action is add_device_action:
            device = self.add_device(node)
            if device is not None:
                self.ui.tree_view.selectionModel().setCurrentIndex(
                    self.model.index_from_node(device),
                    QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows)
        elif action is flash_action:
            self.flash(interface=node.interface,
                       channel=node.channel)
        elif action is pull_raw_log_action:
            self.pull_raw_log(node=node)
        elif action is write_to_epz_action:
            self.write_to_epz(device=node.device)
        elif action is check_compatibility_action:
            self.check_compatibility(device=node.device)
        elif action is blink_action:
            self.blink(bus=node.bus)

    def blink(self, bus):
        bus.flash()

    def check_compatibility(self, device):
        shas = device.shas
        if len(shas) == 0:
            self.compatibility_notification(
                message='No compatible revisions specified for this device.',
                compatible=False
            )
        else:
            try:
                signal = device.nvs.signal_from_names(
                    'SoftwareHash', 'SoftwareHash')
            except epyqlib.nv.NotFoundError as e:
                raise UnsupportedError(
                    'Compatibility check is not supported for this device') from e

            d = device.nvs.protocol.read(signal)
            d.addCallback(self._check_compatibility, device=device)
            d.addErrback(epyqlib.utils.twisted.errbackhook)

    def _check_compatibility(self, value, device):
        sha = '{:07x}'.format(int(value))

        compatible = any(s.startswith(sha) for s in device.shas)
        print('alskflasdfasdfaslkfdasdfjsdjfasdfsd')
        print(sha)
        print(device.shas)

        if compatible:
            message = ('Embedded SHA {} found in compatibility list.'
                       .format(sha))
        else:
            message = ('Embedded SHA {} not found in compatibility list.'
                       .format(sha))

        self.compatibility_notification(
            message=message,
            compatible=compatible
        )

    def compatibility_notification(self, message, compatible):
        if compatible:
            icon = QtWidgets.QMessageBox.Information
        else:
            icon = QtWidgets.QMessageBox.Warning

        epyqlib.utils.qt.dialog(
            parent=self,
            message=message,
            icon=icon,
        )

    def pull_raw_log(self, node):
        bus_node = node.tree_parent
        real_bus = can.interface.Bus(bustype=bus_node.interface,
                                     channel=bus_node.channel,
                                     bitrate=bus_node.bitrate)
        bus = epyqlib.busproxy.BusProxy(bus=real_bus,
                                        auto_disconnect=False)

        d = epyqlib.datalogger.pull_raw_log(
            device=node.device, bus=bus, parent=self)
        d.addBoth(epyqlib.utils.twisted.detour_result, bus.set_bus)

    def add_device(self, bus, device=None):
        if device is None:
            device = load_device(parent=self)

        if device is not None:
            device = epyqlib.devicetree.Device(device=device)

            self.model.add_device(bus, device)
            index = self.model.index_from_node(bus)
            index = self.model.index(index.row(), 0, index.parent())
            self.ui.tree_view.setExpanded(index, True)

        return device

    def flash(self, interface, channel):
        # TODO  CAMPid 9561616237195401795426778
        filters = [
            ('TICOFF Binaries', ['out']),
            ('All Files', ['*'])
        ]
        file = epyqlib.utils.qt.file_dialog(filters, parent=self)

        if file is not None:
            text = textwrap.dedent('''\
            Flashing {file}

            <ol>
              <li> Prepare module for programming
                <ul>
                  <li> Remove all high power from the module </li>
                  <li> Remove 24V control power </li>
                  <li> Install jumper between J1 pin 1 and pin 2 (+24V to Boot Enable), but do not reconnect 24V at this time </li>
                  <li> Ensure that the USB to CAN device is installed and connected to the appropriate CAN network </li>
                  <br>
                  <br>
                  <i> <b>Note:</b> no other devices may be active on the CAN network during programming operations </i>
                  <br>
                  <i> <b>Note:</b> firmware programming occurs at a CAN baud rate of 250kbits/sec.  If any CAN tools are in use during programming, they should be set to a baud rate of 250kbits/sec </i>
                </ul>
              </li>
              <br>
              <li> Apply firmware update via CAN
                <ul>
                  <li> Click OK </li>
                  <li> Reapply 24V control power </li>
                  <li> EPyQ will search for the module and should find it within a second or two </li>
                  <li> EPyQ will initiate clearing of module's flash memory </li>
                  <li> EPyQ will flash the device providing progress status </li>
                  <li> This whole process should take 1-2 minutes </li>
                </ul>
              </li>
              <br>
              <li> Restore System
                <ul>
                  <li> Remove 24V control power </li>
                  <li> Remove Boot Enable jumper </li>
                  <li> Power up module normally </li>
                </ul>
              </li>
            </ol>
            '''.format(file=file))

            accepted = epyqlib.utils.qt.dialog(
                parent=self,
                title='Flashing',
                message=text,
                icon=QtWidgets.QMessageBox.Information,
                rich_text=True,
                cancellable=True,
            )

            if accepted:
                with open(file, 'rb') as f:
                    real_bus = can.interface.Bus(bustype=interface,
                                                 channel=channel,
                                                 bitrate=250000)
                    bus = epyqlib.busproxy.BusProxy(bus=real_bus,
                                                    auto_disconnect=False)

                    progress = epyqlib.utils.qt.progress_dialog(
                        parent=self,
                        cancellable=True
                    )

                    flasher = epyqlib.flash.Flasher(file=f,
                                                    bus=bus,
                                                    progress=progress,
                                                    retries=math.inf,
                                                    parent=self)

                    failed_box = QMessageBox(parent=self)
                    failed_box.setText(textwrap.dedent('''\
                    Flashing failed
                    '''))

                    canceled_box = QMessageBox(parent=self)
                    canceled_box.setText(textwrap.dedent('''\
                    Flashing canceled
                    '''))

                    flasher.done.connect(progress.close)
                    flasher.done.connect(progress.deleteLater)
                    flasher.done.connect(bus.set_bus)

                    completed_format = textwrap.dedent('''\
                    Flashing completed successfully

                    Data time: {:.3f} seconds for {} bytes or {:.0f} bytes/second''')
                    flasher.completed.connect(
                        lambda f=flasher:
                            print(
                                completed_format.format(
                                    f.data_delta_time,
                                    f.download_bytes,
                                    f.download_bytes / f.data_delta_time
                                )
                            )
                    )
                    flasher.completed.connect(
                        lambda f=flasher:
                            QMessageBox.information(
                                self,
                                'EPyQ',
                                completed_format.format(
                                    f.data_delta_time,
                                    f.download_bytes,
                                    f.download_bytes / f.data_delta_time
                                )
                            )
                    )
                    flasher.failed.connect(failed_box.exec)
                    flasher.canceled.connect(canceled_box.exec)
                    flasher.done.connect(bus.set_bus)

                    flasher.flash()

    def remove_device(self, device):
        self.ui.tree_view.clearSelection()
        self.model.remove_device(device)

    def write_to_epz(self, device):
        print(device.referenced_files)

        filters = [
            ('EPZ', ['epz']),
            ('All Files', ['*'])
        ]
        filename = epyqlib.utils.qt.file_dialog(
            filters, save=True, parent=self)

        if filename is not None:
            epyqlib.utils.general.write_device_to_zip(
                zip_path=filename,
                epc_dir='',
                referenced_files=device.referenced_files,
                code=epyqlib.utils.qt.get_code()
            )

    def setModel(self, model):
        self.model = model
        self.ui.tree_view.setModel(model)

        self.ui.tree_view.header().setStretchLastSection(False)

        for i, resize in enumerate(self.resize_columns):
            if resize:
                self.ui.tree_view.header().setSectionResizeMode(
                    i, QtWidgets.QHeaderView.ResizeToContents)

        self.ui.tree_view.setItemDelegateForColumn(
            epyqlib.devicetree.Columns.indexes.bitrate,
            epyqlib.delegates.ByFunction(model=model, parent=self)
        )

        self.ui.tree_view.selectionModel().currentChanged.connect(
            self._current_changed)

        widths = [self.ui.tree_view.columnWidth(i)
                  for i in epyqlib.devicetree.Columns.indexes]
        width = sum(widths)
        width += 2 * self.ui.tree_view.frameWidth()

        self.ui.tree_view.setMinimumWidth(1.25 * width)

        self.ui.tree_view.header().setSectionResizeMode(
            epyqlib.devicetree.Columns.indexes.name,
            QHeaderView.Stretch
        )


if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure



#!/usr/bin/env python
"""
Read TI COFF file into a python object
"""


from __future__ import print_function

from builtins import bytes, object, range
from array import array
from collections import namedtuple
from optparse import OptionParser
from struct import unpack, calcsize


# See file COPYING in this source tree
# See file ticoff.asi_license.txt in this source tree
__copyright__ = ('\n'.join([
    'Copyright 2011 "Eliot Blennerhassett" <eblennerhassett@audioscience.com>',
    # https://gist.github.com/eliotb/1073231

    'Copyright 2016, EPC Power Corp.'
    # Added Python3 support and a few tweaks
]))
__license__ = 'GPLv2+'


def read_struct(file, format):
    '''read struct data formatted according to format'''
    datalen = calcsize(format)
    data = file.read(datalen)
    if len(data) != datalen:
        raise EOFError
    ret = unpack(format, data)
    if len(ret) < 2:
        ret = ret[0]
    return ret


def read_cstr(file):
    '''read zero terminated c string from file'''
    output = ""
    while True:
        char = file.read(1)
        if len(char) == 0:
            raise RuntimeError ("EOF while reading cstr")
        char = chr(bytes(char)[0])
        if char == '\0':
            break
        output += char
    return output

# From spraao8.pdf table 7
class Section(namedtuple('Section',
    'name, virt_size, virt_addr, raw_data_size, raw_data_ptr, relocation_ptr,' +
    'linenum_ptr, linenum_count, reloc_count, flags, reserved, mem_page, data')):

    __slots__ = () # prevent creation of instance dictionaries
    section_fmt = '<8s9L2H'
    section_flags = [
        (0x00000001, 'STYP_DSECT'), (0x00000002, 'STYP_NOLOAD'),
        (0x00000004, 'STYP_GROUPED'), (0x00000008, 'STYP_PAD'),
        (0x00000010, 'STYP_COPY'), (0x00000020, 'STYP_TEXT'),
        (0x00000040, 'STYP_DATA'), (0x00000080, 'STYP_BSS'),
        (0x00000100, 'STYP_100'), (0x00000200, 'STYP_200'),
        (0x00000200, 'STYP_400'), (0x00000800, 'STYP_800'),
        (0x00001000, 'STYP_BLOCK'), (0x00002000, 'STYP_PASS'),
        (0x00004000, 'STYP_CLINK'),(0x00008000, 'STYP_VECTOR'),
        (0x00010000, 'STYP_PADDED'),
    ]

    @property
    def is_zero_filled(self):
        return (self.data is not None) and not any([d != '\0' for d in self.data])

    def fill_value(self):
        if self.data is  None:
            return None

        if len(self.data) % 4:
            return None # only handle 32 bit fill

        a = array('I', self.data)
        if all([d == a[0] for d in a]):
            return a[0]
        else:
            return None

    @property
    def is_loadable(self):
        return not(self.flags & 0x00000010 or self.data is None)

    def __str__(self):
        s = 'Section %s' % self.name
        if self.flags:
            s += ' Flags ='
            for f in Section.section_flags:
                if self.flags & f[0]:
                    s = s + ' ' + f[1]
        for fv in self._asdict().items():
            if fv[0] == 'data' or fv[0] == 'name': continue
            s = s + '\n\t%s = 0x%X' % fv
        return s

class Symbol(namedtuple('Symbol', 'name, value, section_number, reserved,'
                                  'storage_class, number_of_aux_entries')):

    __slots__ = () # prevent creation of instance dictionaries
    symbol_fmt = '<8sLhHcc'
    symbol_auxiliary_format = '<LHH10c'
    symbol_flags = [
        (0, 'C_NULL'),
        (1, 'C_AUTO'),
        (2, 'C_EXT'),
        (3, 'C_STAT'),
        (4, 'C_REG'),
        (5, 'C_EXTREF'),
        (6, 'C_LABEL'),
        (7, 'C_ULABEL'),
        (8, 'C_MOS'),
        (9, 'C_ARG'),
        (10, 'C_STRTAG'),
        (11, 'C_MOU'),
        (12, 'C_UNTAG'),
        (13, 'C_TPDEF'),
        (14, 'C_USTATIC'),
        (15, 'C_ENTAG'),
        (16, 'C_MOE'),
        (17, 'C_REGPARM'),
        (18, 'C_FIELD'),
        (19, 'C_UEXT'),
        (20, 'C_STATLAB'),
        (21, 'C_EXTLAB'),
        (27, 'C_VARARG'),
        (100, 'C_BLOCK'),
        (101, 'C_FCN'),
        (102, 'C_EOS'),
        (103, 'C_FILE'),
        (104, 'C_LINE')
    ]

class Coff(object):
    Header = namedtuple('Header',
        'machine_type, section_count, time_date, symbol_table_ptr, ' +
        'symbol_count, optional_header_size, flags, target_id')
    header_fmt = '<2H3L3H'

    OptionalHeader = namedtuple('OptionalHeader',
        'magic, version, exe_size, init_data_size, uninit_data_size, entry_point, exe_start, init_start')
    optheader_fmt = '<2H6L'

    def __init__(self, filename=None):
        self.header = None
        self.optheader = None
        self.sections = []
        if filename is not None:
            self.from_file(filename)

    def from_file(self, name):
        with open(name, 'rb') as f:
            self.from_stream(f)

    def from_stream(self, f):
        self.header = self.Header(*read_struct(f, self.header_fmt))
        self.optheader = self.OptionalHeader(*read_struct(f, self.optheader_fmt))
        self.sections = []
        for i in range(self.header.section_count):
            section = Section(*read_struct(f, Section.section_fmt), data=None)
            section = section._replace(name=self.symname(f, section.name))
            if section.raw_data_ptr and section.raw_data_size:
                here = f.tell()
                f.seek(section.raw_data_ptr)
                # TODO: `2 *` is hard coded to handle the 2-bytes per
                #       address scenario.  This should obviously be
                #       detected somehow, unless it is always correct.
                data = bytes(f.read(2 * section.raw_data_size))
                f.seek(here)
                section = section._replace(data=data)
            self.sections.append(section)

        self.symbols = []
        f.seek(self.header.symbol_table_ptr)
        for i in range(self.header.symbol_count):
            symbol = Symbol(*read_struct(f, Symbol.symbol_fmt))
            try:
                symbol = symbol._replace(name=self.symname(f, symbol.name))
            except UnicodeDecodeError:
                # TODO: not sure what to do with these
                pass
            symbol = symbol._replace(
                number_of_aux_entries=symbol.number_of_aux_entries[0])
            try:
                symbol = symbol._replace(
                    storage_class=next(f for f in symbol.symbol_flags
                                   if symbol.storage_class[0] == f[0]))
            except StopIteration:
                print('bad {}'.format(symbol))
            self.symbols.append(symbol)

            # TODO: this breaks by hitting EOF but it seems we would need
            #       to read the aux entries somewhere
            # if symbol.number_of_aux_entries:
            #     read_struct(f, Symbol.symbol_auxiliary_format)

        stack_section_number = next(i for i, s in enumerate(self.sections)
                                    if s.name == '.stack')
        self.variables = [str(s) for s in self.symbols
                          if isinstance(s.name, str)
                          and s.name.startswith('_')
                          and s.section_number == stack_section_number]

        self.entry_point = self.optheader.entry_point
        self.sections.sort(key=lambda s: s.virt_addr)

    @property
    def loadable_sections(self):
        return [s for s in self.sections if s.is_loadable]

    def string_table_entry (self, f, offset):
        here = f.tell()
        seek = self.header.symbol_table_ptr + self.header.symbol_count * 18 + offset
        f.seek(seek)
        s = read_cstr(f)
        f.seek(here)
        return s

    def symname(self, f, value):
        parts = unpack("<2L", value)
        if parts[0] == 0:
            return self.string_table_entry(f, parts[1])
        else:
            return str(value.decode('ascii').rstrip('\0'))

    def __str__(self):
        s = 'TI coff file'
        if self.header is not None:
            for fv in self.header._asdict().items():
                s = s + '\n\t%s = 0x%X' % fv
        if self.optheader is not None:
            s += '\nOptional Header'
            for fv in self.optheader._asdict().items():
                s = s + '\n\t%s = 0x%X' % fv
        return s

if __name__ == '__main__':
    parser = OptionParser()
    parser.add_option("-d", "--dump_data",
                      action="store_true", dest="dump", default=False,
                      help="dump section data")

    options, args = parser.parse_args()

    c = Coff(args[0])
    print(c)

    for section in c.sections:
        print(section)
        if section.data is not None and options.dump:
            print(repr(section))




# This file has been generated


build_system = 'AppVeyor'
build_id = '30440598'
build_number = '453'
build_version = '1.0.453'
job_id = 'uxgqfytggh822ud5'
job_url = 'https://ci.appveyor.com/project/KyleAltendorf/stlib/build/job/uxgqfytggh822ud5'



import pathlib

from PyQt5 import QtCore, QtWidgets
import twisted.internet.defer

import epyqlib.scriptingview_ui
import epyqlib.utils.qt
import epyqlib.utils.twisted


# See file COPYING in this source tree
__copyright__ = 'Copyright 2017, EPC Power Corp.'
__license__ = 'GPLv2+'


class ScriptAlreadyActiveError(epyqlib.utils.general.ExpectedException):
    def expected_message(self):
        return 'Script already active.'


class ScriptNotActiveError(epyqlib.utils.general.ExpectedException):
    def expected_message(self):
        return 'Script is not active.'


class ScriptAlreadyPausedError(epyqlib.utils.general.ExpectedException):
    def expected_message(self):
        return 'Script already paused.'


class ScriptNotPausedError(epyqlib.utils.general.ExpectedException):
    def expected_message(self):
        return 'Script is not paused.'


def cancelled_handler(error):
    if isinstance(error.value, twisted.internet.defer.CancelledError):
        epyqlib.utils.qt.raw_exception_message_box(
            brief='Script cancelled by user request.',
            extended=error.getTraceback(),
        )

        return None

    return error


class ScriptingView(QtWidgets.QWidget):
    closing = QtCore.pyqtSignal()

    def __init__(self, parent=None, in_designer=False):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        self.ui = epyqlib.scriptingview_ui.Ui_Form()
        self.ui.setupUi(self)

        self.ui.load_button.clicked.connect(self.load)
        self.ui.save_button.clicked.connect(self.save)

        self.ui.run_button.clicked.connect(self.run)
        self.ui.loop_button.clicked.connect(self.loop)
        self.ui.stop_button.clicked.connect(self.stop)

        self.ui.pause_button.clicked.connect(self.pause)
        self.ui.continue_button.clicked.connect(self.unpause)

        self.model = None
        self.model_connections = []

        with open(pathlib.Path(__file__).parents[0] / 'scripting.csv') as f:
            self.ui.csv_edit.setPlaceholderText(f.read())

        self.sequence = None
        self.update_buttons()

    def closeEvent(self, event):
        result = super().closeEvent(event)
        self.closing.emit()
        return result

    def update_buttons(self):
        active = self.sequence is not None

        self.ui.run_button.setEnabled(not active)
        self.ui.loop_button.setEnabled(not active)
        self.ui.stop_button.setEnabled(active)

        self.ui.pause_button.setEnabled(active and not self.sequence.paused)
        self.ui.continue_button.setEnabled(active and self.sequence.paused)

    def set_model(self, model):
        for connection in self.model_connections:
            connection.disconnect()

        self.model = model

    def load(self):
        filters = [
            ('CSV', ['csv']),
            ('All Files', ['*'])
        ]
        filename = epyqlib.utils.qt.file_dialog(
            filters,
            parent=self,
        )

        if filename is None:
            return

        with open(filename) as f:
            self.ui.csv_edit.setPlainText(f.read())

    def save(self):
        filters = [
            ('CSV', ['csv']),
            ('All Files', ['*'])
        ]
        filename = epyqlib.utils.qt.file_dialog(
            filters,
            save=True,
            parent=self.ui,
        )

        if filename is None:
            return

        with open(filename, 'w') as f:
            text = self.ui.csv_edit.toPlainText()
            f.write(text)
            if text[-1] != '\n':
                f.write('\n')

    def stop(self):
        if self.sequence is None:
            return

        self.sequence.cancel()
        self.sequence = None
        self.update_buttons()

    def errback(self):
        self.sequence = None
        self.update_buttons()

    def run(self):
        return self._run(loop=False)

    def loop(self):
        return self._run(loop=True)

    def _run(self, loop=False):
        if self.sequence is not None:
            raise ScriptAlreadyActiveError()

        self.sequence = self.model.run_s(
            event_string=self.ui.csv_edit.toPlainText(),
            pause=self.pause,
            loop=loop,
        )
        self.update_buttons()

        self.sequence.run_deferred.addBoth(
            epyqlib.utils.twisted.detour_result,
            self.errback,
        )
        self.sequence.run_deferred.addErrback(
            epyqlib.utils.twisted.catch_expected,
        )
        self.sequence.run_deferred.addErrback(cancelled_handler)
        self.sequence.run_deferred.addErrback(epyqlib.utils.twisted.errbackhook)

    def pause(self):
        if self.sequence is None:
            raise ScriptNotActiveError()

        if self.sequence.paused:
            raise ScriptAlreadyPausedError()

        self.sequence.pause()
        self.update_buttons()

    def unpause(self):
        if self.sequence is None:
            raise ScriptNotActiveError()

        if not self.sequence.paused:
            raise ScriptNotPausedError()

        self.sequence.unpause()
        self.update_buttons()



import enum

import attr
import graham
import marshmallow

import epyqlib.attrsmodel
import epyqlib.treenode


@graham.schemify(tag='node')
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Node(epyqlib.treenode.TreeNode):
    name = epyqlib.attrsmodel.create_name_attribute()
    epyqlib.attrsmodel.attrib(
        attribute=name,

    )

    node = attr.ib(default=None)
    epyqlib.attrsmodel.attrib(
        attribute=node,
        no_column=True,
    )

    children = attr.ib(
        default=attr.Factory(list),
    )

    uuid = epyqlib.attrsmodel.attr_uuid(no_column=True)

    def __attrs_post_init__(self):
        super().__init__()

    @classmethod
    def build(cls, name, node, local_results=[], child_results=[]):
        results = [
            Result(node=node, message=result)
            for result in local_results
        ]

        node = cls(
            name=name,
            node=node,
        )

        for child in child_results + results:
            node.append_child(child)

        return node

    def can_delete(self, node=None):
        return False

    def check(self, models):
        return None

    def can_drop_on(self, node):
        return False

    all_addable_types = epyqlib.attrsmodel.empty_all_addable_types
    addable_types = epyqlib.attrsmodel.empty_addable_types
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move


@enum.unique
class ResultSeverity(enum.Enum):
    information = 0
    warning = 1
    error = 2


@graham.schemify(tag='result')
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Result(epyqlib.treenode.TreeNode):
    node = attr.ib(default=None)
    epyqlib.attrsmodel.attrib(
        attribute=node,
        no_column=True,
    )

    severity = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    epyqlib.attrsmodel.attrib(
        attribute=severity,
        data_display=epyqlib.attrsmodel.name_from_enumerator,
    )
    message = epyqlib.attrsmodel.create_str_attribute()

    children = attr.ib(
        default=attr.Factory(list),
    )

    uuid = epyqlib.attrsmodel.attr_uuid(no_column=True)

    def __attrs_post_init__(self):
        super().__init__()

    def can_delete(self, node=None):
        return False

    def check(self, models):
        return None

    def can_drop_on(self, node):
        return False

    all_addable_types = epyqlib.attrsmodel.empty_all_addable_types
    addable_types = epyqlib.attrsmodel.empty_addable_types
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move


Root = epyqlib.attrsmodel.Root(
    default_name='Check Results',
    valid_types=(
        Node,
        Result,
    ),
)


types = epyqlib.attrsmodel.Types(
    types=(
        Root,
        Node,
        Result,
    ),
)


# TODO: CAMPid 943896754217967154269254167
def merge(name, *types):
    return tuple((x, name) for x in types)


columns = epyqlib.attrsmodel.columns(
    merge('name', Node),
    merge('severity', Result),
    merge('message', Result),
)


# TODO: CAMPid 075454679961754906124539691347967
@attr.s
class ReferencedUuidNotifier:
    changed = epyqlib.utils.qt.Signal('PyQt_PyObject')

    view = attr.ib(default=None)
    selection_model = attr.ib(default=None)

    def __attrs_post_init__(self):
        if self.view is not None:
            self.set_view(self.view)

    def set_view(self, view):
        self.disconnect_view()

        self.view = view
        self.selection_model = self.view.selectionModel()
        if self.selection_model is not None:
            self.selection_model.currentChanged.connect(
                self.current_changed,
            )

    def disconnect_view(self):
        if self.selection_model is not None:
            self.selection_model.currentChanged.disconnect(
                self.current_changed,
            )
        self.view = None
        self.selection_model = None

    def current_changed(self, current, previous):
        if not current.isValid():
            return

        index = epyqlib.utils.qt.resolve_index_to_model(
            index=current,
        )
        model = index.data(epyqlib.utils.qt.UserRoles.attrs_model)
        node = model.node_from_index(index)
        self.changed.emit(node.node.uuid)



import collections
import decimal
import json
import os
import pathlib
import shutil
import subprocess
import tempfile
import zipfile

import attr
import canmatrix.formats

import epyqlib.device


class InvalidAutoParametersDeviceError(Exception):
    pass


class MultipleFoundError(Exception):
    pass


raw_template = pathlib.Path(epyqlib.autodevice.__file__).with_name('template')


def create_template_archive(path):
    with zipfile.ZipFile(path, 'w') as z:
        for template_file in raw_template.iterdir():
            z.write(
                filename=template_file,
                arcname=pathlib.Path('autodevice')/template_file.name,
            )


@attr.s
class AccessInput:
    node = attr.ib()
    description = attr.ib()
    secret = attr.ib()
    value = attr.ib(default=None)


@attr.s
class Builder:
    access_parameters = attr.ib(default=None)
    required_serial_number = attr.ib(default=None)
    archive_code = attr.ib()
    _template = attr.ib(default=None, init=False)
    _template_path = attr.ib(default=None, init=False)
    _value_set = attr.ib(default=None, init=False)
    _target = attr.ib(default=None, init=False)
    _original_raw_dict = attr.ib(default=None, init=False)
    _temporary_directory = attr.ib(default=None, init=False)

    @archive_code.default
    def _(self):
        code = epyqlib.utils.qt.get_code()
        if code is not None:
            code = code.decode('ascii')

        return code

    def set_access_level_names(self, password_name, access_level_name):
        self.access_parameters = (
            AccessInput(
                node=password_name,
                description='Elevated Access Code',
                secret=True,
            ),
            AccessInput(
                node=access_level_name,
                description='Elevated Access Level',
                secret=False,
            ),
        )

    def set_template(self, path, archive=False):
        path = pathlib.Path(path)

        if archive:
            self._temporary_directory = tempfile.TemporaryDirectory()
            with zipfile.ZipFile(path) as z:
                z.extractall(self._temporary_directory.name)

            self._template_path, = pathlib.Path(
                self._temporary_directory.name,
            ).glob('**/*.epc')
        else:
            self._template_path = path

        self._template = epyqlib.device.Device(
            file=os.fspath(self._template_path),
            only_for_files=True,
        )

        auto_value_set_key = 'auto_value_set'
        if auto_value_set_key not in self._template.raw_dict:
            raise InvalidAutoParametersDeviceError(
                'Key {} not found'.format(auto_value_set_key)
            )

    def set_original_raw_dict(self, original_raw_dict):
        self._original_raw_dict = original_raw_dict

    def load_pmvs(self, path):
        self._value_set = epyqlib.pm.valuesetmodel.loadp(path)

    def load_epp(self, parameters, can, can_suffix):
        matrix, = canmatrix.formats.load(
            can,
            import_type=can_suffix[1:],
        ).values()
        neo = epyqlib.canneo.Neo(
            matrix=matrix,
            frame_class=epyqlib.nv.Frame,
            signal_class=epyqlib.nv.Nv,
            strip_summary=False,
        )
        nvs = epyqlib.nv.Nvs(
            neo=neo,
            configuration=self._original_raw_dict['nv_configuration'],
        )
        parameters = json.load(parameters)
        nvs.from_dict(parameters)
        self._value_set = nvs.to_value_set(include_secrets=True)

    def load_epp_paths(self, parameter_path, can_path):
        can_suffix = pathlib.Path(can_path).suffix
        with open(can_path, 'rb') as can:
            with open(parameter_path) as parameters:
                self.load_epp(
                    can=can,
                    can_suffix=can_suffix,
                    parameters=parameters,
                )

    def get_or_create_parameter(self, name):
        try:
            nodes = self._value_set.model.root.nodes_by_attribute(
                attribute_value=name,
                attribute_name='name',
            )
        except epyqlib.treenode.NotFoundError:
            node = epyqlib.pm.valuesetmodel.Parameter(
                name=name,
            )
            self._value_set.model.root.append_child(node)
        else:
            try:
                node, = nodes
            except ValueError as e:
                raise MultipleFoundError(
                    'Found multiple nodes but expected only one when '
                    'searching for {}'.format(repr(name))
                )

        return node

    def set_target(self, path):
        self._target = pathlib.Path(path)

    def create(self, can_contents):
        for access_input in self.access_parameters:
            access_input.node = self.get_or_create_parameter(access_input.node)

        with tempfile.TemporaryDirectory() as temporary_directory:
            temporary_directory = pathlib.Path(temporary_directory)
            directory_path = temporary_directory / self._target.stem
            directory_path.mkdir()

            for access_input in self.access_parameters:
                node = self.get_or_create_parameter(name=access_input.node.name)

                node.value = access_input.value

            self._value_set.path = directory_path / self._template.raw_dict[
                'auto_value_set'
            ]
            self._value_set.save()

            can_path = self._template.raw_dict['can_path']

            for file_name in self._template.referenced_files:
                skips = (
                    can_path,
                    self._template.raw_dict['auto_value_set'],
                )
                if file_name in skips:
                    continue

                file_path = pathlib.Path(file_name)

                shutil.copy(
                    self._template_path.parent/file_path.name,
                    directory_path,
                )

            with open(self._template_path) as f:
                raw_dict = json.load(
                    f,
                    object_pairs_hook=collections.OrderedDict,
                )

            keys_to_copy = (
                'can_configuration',
                'nv_configuration',
                'node_id_type',
                'access_level_path',
                'access_password_path',
                'nv_meta_enum',
            )
            for key in keys_to_copy:
                if key in self._original_raw_dict:
                    raw_dict[key] = self._original_raw_dict[key]

            if isinstance(self.required_serial_number, decimal.Decimal):
                raw_dict['required_serial_number'] = int(
                    self.required_serial_number
                )
            else:
                raw_dict['required_serial_number'] = (
                    self.required_serial_number
                )

            can_path = directory_path / can_path
            with open(can_path, 'wb') as f:
                f.write(can_contents)

            matrix, = canmatrix.formats.loadp(os.fspath(can_path)).values()
            neo = epyqlib.canneo.Neo(
                matrix=matrix,
                frame_class=epyqlib.nv.Frame,
                signal_class=epyqlib.nv.Nv,
                strip_summary=False,
            )
            nvs = epyqlib.nv.Nvs(
                neo=neo,
                configuration=self._original_raw_dict['nv_configuration'],
            )

            serial_nv, = [
                nv
                for nv in nvs.all_nv()
                if (
                    'serial' in nv.name.casefold()
                    and 'number' in nv.name.casefold()
                )
            ]

            raw_dict['serial_number_names'] = (
                serial_nv.frame.mux_name,
                serial_nv.name,
            )

            target_epc_name = (
                directory_path / 'auto_parameters.epc'
            )
            with open(target_epc_name, 'w') as f:
                json.dump(raw_dict, f, indent=4)

            backup_path = None
            if self._target.exists():
                backup_path = temporary_directory / self._target.name
                shutil.move(self._target, backup_path)

            try:
                password_option = []
                if len(self.archive_code) > 0:
                    password_option = ['-p{}'.format(self.archive_code)]

                paths = (
                    pathlib.Path('7z'),
                    (
                        pathlib.Path('C:')
                        / os.sep
                        / 'Program Files'
                        / '7-Zip'
                        / '7z.exe'
                    ),
                    (
                        pathlib.Path('C:')
                        / os.sep
                        / 'Program Files (x86)'
                        / '7-Zip'
                        / '7z.exe'
                    ),
                )
                for path in paths:
                    try:
                        subprocess.run(
                            [
                                os.fspath(path),
                                'a',
                                '-tzip',
                                os.fspath(self._target),
                                os.fspath(directory_path),
                                *password_option,
                            ],
                            check=True,
                        )
                    except FileNotFoundError:
                        continue
                    else:
                        break
                else:
                    raise Exception(
                        'Unable to find 7z binary as any of: {}'.format(
                            paths,
                        )
                    )
            except Exception as e:
                if backup_path is not None:
                    shutil.move(backup_path, self._target)

                raise e

        if self._temporary_directory is not None:
            self._temporary_directory.cleanup()



import json

import attr
import ccstudiodss.api
import ccstudiodss.cli
import click
import javabridge

import epyqlib.cmemoryparser
import epyqlib.variableselectionmodel


project_name = 'epyqlib'


@click.group()
def main():
    pass


@main.command()
@ccstudiodss.cli.create_binary_option(project_name=project_name)
@ccstudiodss.cli.create_ccxml_option(project_name=project_name)
@ccstudiodss.cli.ccs_base_path_option
def dss(binary, ccxml, ccs_base_path):
    ccstudiodss.api.add_jars(base_path=ccs_base_path)

    model = epyqlib.variableselectionmodel.VariableModel(
        nvs=None,
        nv_model=None,
        bus=None,
    )

    binary_info = epyqlib.cmemoryparser.process_file(binary)
    model.update_from_loaded_binary_without_threads(binary_info=binary_info)

    try:
        with ccstudiodss.api.Session(ccxml=ccxml) as session:
            session.load(binary=binary, timeout=10000)
            session.run()

            traverser = Traverser(session=session)

            try:
                model.root.traverse(call_this=traverser, internal_nodes=True)
            except Exception as e:
                print()

            for comparison in traverser.failures():
                print('FAILED:', comparison.node.qualified_name(), comparison)

            print()
    except Exception as e:
        print()


@main.command(name='json')
@ccstudiodss.cli.create_binary_option(project_name=project_name)
@click.option('--json', 'json_file', type=click.File('r'), required=True)
def json_command(binary, json_file):
    model = epyqlib.variableselectionmodel.VariableModel(
        nvs=None,
        nv_model=None,
        bus=None,
    )

    binary_info = epyqlib.cmemoryparser.process_file(binary)
    model.update_from_loaded_binary_without_threads(binary_info=binary_info)

    loaded = json.load(json_file)

    try:
        traverser = JsonTraverser(name_to_addresses=dict(loaded))

        try:
            model.root.traverse(call_this=traverser, internal_nodes=True)
        except Exception as e:
            print()

        for comparison in traverser.failures():
            print('FAILED:', comparison.node.qualified_name(), comparison)

        print()
    except Exception as e:
        print()


@attr.s
class Comparison:
    node = attr.ib(default=None)
    dss_address = attr.ib(default=None)
    exception = attr.ib(default=None)

    def matches(self):
        if self.node is None:
            return False

        # if isinstance(self.node.variable, epyqlib.cmemoryparser.Variable):
        #     address = self.node.variable.address
        # elif isinstance(self.node.variable, epyqlib.cmemoryparser.StructMember):
        #

        address = int(self.node.fields.address[2:], 16)

        return address == self.dss_address


@attr.s
class Traverser:
    session = attr.ib()
    collected = attr.ib(factory=list)

    def __call__(self, node, payload):
        if node.tree_parent is None:
            return

        # if node.tree_parent.tree_parent is not None:
        #     print()

        print(node.qualified_name())

        try:
            dss_address = self.session.debug_session.expression.evaluate(
                '&' + node.qualified_name()
            )
        except javabridge.JavaException as e:
            comparison = Comparison(node=node, exception=e)
        else:
            comparison = Comparison(node=node, dss_address=dss_address)

        self.collected.append(comparison)

    def failures(self):
        return [
            comparison
            for comparison in self.collected
            if not comparison.matches()
        ]


@attr.s
class JsonTraverser:
    name_to_addresses = attr.ib()
    collected = attr.ib(factory=list)

    def __call__(self, node, payload):
        if node.tree_parent is None:
            return

        # if node.tree_parent.tree_parent is not None:
        #     print()

        print(node.qualified_name())

        try:
            dss_address = self.name_to_addresses[node.qualified_name()]
        except KeyError as e:
            comparison = Comparison(node=node, exception=e)
        else:
            comparison = Comparison(node=node, dss_address=dss_address)

        self.collected.append(comparison)

    def failures(self):
        return [
            comparison
            for comparison in self.collected
            if not comparison.matches()
        ]



import epyqlib.attrsmodel
import epyqlib.checkresultmodel


TestAttrsModel = epyqlib.attrsmodel.build_tests(
    types=epyqlib.checkresultmodel.types,
    root_type=epyqlib.checkresultmodel.Root,
    columns=epyqlib.checkresultmodel.columns,
)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'compoundtoggle.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(152, 169)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Form.sizePolicy().hasHeightForWidth())
        Form.setSizePolicy(sizePolicy)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(Form)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.box = QtWidgets.QGroupBox(Form)
        self.box.setObjectName("box")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.box)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.command = Toggle(self.box)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.command.sizePolicy().hasHeightForWidth())
        self.command.setSizePolicy(sizePolicy)
        self.command.setProperty("label_visible", False)
        self.command.setProperty("tx", True)
        self.command.setProperty("value_labels_visible", False)
        self.command.setObjectName("command")
        self.horizontalLayout.addWidget(self.command)
        self.verticalLayout_12 = QtWidgets.QVBoxLayout()
        self.verticalLayout_12.setObjectName("verticalLayout_12")
        self.status_on = Led(self.box)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.status_on.sizePolicy().hasHeightForWidth())
        self.status_on.setSizePolicy(sizePolicy)
        self.status_on.setProperty("label_override", "")
        self.status_on.setProperty("label_visible", True)
        self.status_on.setProperty("label_from_enumeration", True)
        self.status_on.setObjectName("status_on")
        self.verticalLayout_12.addWidget(self.status_on)
        self.status_off = Led(self.box)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.status_off.sizePolicy().hasHeightForWidth())
        self.status_off.setSizePolicy(sizePolicy)
        self.status_off.setProperty("label_override", "")
        self.status_off.setProperty("label_visible", True)
        self.status_off.setProperty("label_from_enumeration", True)
        self.status_off.setProperty("on_value", 0)
        self.status_off.setObjectName("status_off")
        self.verticalLayout_12.addWidget(self.status_off)
        self.horizontalLayout.addLayout(self.verticalLayout_12)
        self.horizontalLayout_2.addWidget(self.box)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.box.setTitle(_translate("Form", "GroupBox"))
from epyqlib.widgets.led import Led
from epyqlib.widgets.toggle import Toggle



#!/usr/bin/env python3

import argparse
import collections
import io
import os
import platform
import requests
import shutil
import sys
import tempfile
import textwrap
import zipfile


Activate = collections.namedtuple(
    'Activate',
    [
        'windows',
        'set_format',
        'path_separator',
        'pwd',
        'comment_marker',
        'insert_before'
    ]
)

activate_scripts = {
    'activate': Activate(
        windows=False,
        set_format='export {name}={value}',
        path_separator=':',
        pwd='$(pwd)',
        comment_marker='#',
        insert_before=None
    ),
    'activate.bat': Activate(
        windows=True,
        set_format='set {name}={value}',
        path_separator=';',
        pwd='%cd%',
        comment_marker='REM',
        insert_before = None
    ),
    'activate.ps1': Activate(
        windows=True,
        set_format='${name} = "{value}"',
        path_separator=';',
        pwd="$((Get-Location).Path)",
        comment_marker='#',
        insert_before='# SIG'
    ),
}


def download_zips(directory):
    src_zips = {
        'fontawesome':
            'https://github.com/FortAwesome/Font-Awesome/archive/v4.6.3.zip',
        'metropolis':
            'https://github.com/chrismsimpson/Metropolis/archive/16882c2c2cb58405fd6a7d6a932a1dfc573b6813.zip'
    }

    os.makedirs(directory, exist_ok=True)

    for name, url in src_zips.items():
        response = requests.get(url)

        zip_data = io.BytesIO()
        zip_data.write(response.content)
        zip_file = zipfile.ZipFile(zip_data)
        zip_dir = os.path.split(zip_file.namelist()[0])[0]

        with tempfile.TemporaryDirectory() as td:
            zip_file.extractall(path=td)

            zip_path = os.path.join(td, zip_dir)

            destination = os.path.join(directory, name)
            if os.path.exists(destination):
                raise FileExistsError(
                    '`{}` already exists while extracting Zip'.format(destination))

            shutil.move(zip_path, destination)


def write_patch_notice(file, comment_marker):
    file.write(textwrap.dedent('''\


    {c}
    {c} ==== Patch below added to support EPyQ

    '''.format(c=comment_marker)))


def patch_activate(bin):
    path_variables = collections.OrderedDict([
        ('PYQTDESIGNERPATH', [
            ['sub', 'epyqlib', 'epyqlib'],
            ['sub', 'epyqlib', 'epyqlib', 'widgets'],
        ])
    ])

    for name, script in activate_scripts.items():
        path = os.path.join(bin, name)

        if os.path.exists(path):
            set_commands = []

            for variable, paths in path_variables.items():
                paths = [os.path.join(script.pwd, *p) for p in paths]

                command = script.set_format.format(
                    name=variable,
                    value=script.path_separator.join(paths)
                )
                set_commands.append(command)

            if script.insert_before is not None:
                with open(path, 'r') as f:
                    contents = f.read()
                with open(path, 'w') as f:
                    written = False
                    for line in contents.splitlines():
                        if line.startswith(script.insert_before) and not written:
                            written = True
                            write_patch_notice(f, script.comment_marker)

                            for command in set_commands:
                                f.write(command + '\n')
                            f.write('\n\n')

                        f.write(line + '\n')
            else:
                with open(path, 'a') as f:
                    write_patch_notice(f, script.comment_marker)

                    for command in set_commands:
                        f.write(command + '\n')


def write_activate_shortcuts(root, bin):
    for name, script in activate_scripts.items():
        caller = os.path.join(root, name)
        target = os.path.join(bin, name)
        if os.path.isfile(target):
            with open(caller, 'w') as f:
                if script.windows:
                    target = '"{}"'.format(target.replace('\\', '/'))
                else:
                    target = 'source "{}"'.format(target)
                f.write(target + '\n')


def copy_designer_files(root):
    files = ['designer', 'designer.bat', 'designer.ps1', 'designer.vbs']

    for file in files:
        shutil.copy(os.path.join(os.path.dirname(__file__), '..', file),
                    os.path.join(root, file))


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument('--source', '--src', '-s', required=True)
    parser.add_argument('--bin', '-b', required=True)
    parser.add_argument('--root', '-r', required=True)
    parser.add_argument('--for-test', action='store_true')

    return parser.parse_args(args)


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    args = parse_args(args=args)

    download_zips(directory=args.source)
    patch_activate(bin=args.bin)
    if not args.for_test:
        write_activate_shortcuts(root=args.root, bin=args.bin)
        copy_designer_files(root=args.root)
        if platform.system() != 'Windows':
            backup_cwd = os.getcwd()
            os.chdir(os.path.dirname(args.bin))
            os.symlink(os.path.basename(args.bin), 'Scripts')
            os.chdir(backup_cwd)


if __name__ == '__main__':
    sys.exit(main())



import can
from canmatrix import canmatrix
import copy
import decimal
import epyqlib.utils.general
import functools
import itertools
import locale
import logging
import math
from PyQt5.QtCore import (QObject, QTimer, Qt)
import re
import struct
import sys
import uuid

import epyqlib.utils.qt
import epyqlib.utils.units

# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class OutOfRangeError(ValueError):
    pass


class UnableToPackError(Exception):
    pass


class NotFoundError(Exception):
    pass


nothing = object()


def strip_uuid_from_comment(comment):
    match = re.search(r'<uuid:([a-z0-9-]+)>', comment)

    if match is None:
        return comment, None

    uuid_object = uuid.UUID(match[1])

    stripped_comment = comment.replace(match[0], '').strip()

    return stripped_comment, uuid_object


@functools.lru_cache(4096)
def pack_bitstring(length, is_float, value, signed):
    if is_float:
        # TODO: CAMPid 097897541967932453154321546542175421549
        types = {
            32: '>f',
            64: '>d'
        }

        float_type = types.get(length)

        if float_type is None:
            raise Exception(
                'float type only supports lengths in [{}]'.
                    format(', '.join([str(t) for t in types.keys()]))
            )

        x, = struct.pack(float_type, value)

        bitstring = ''.join('{:08b}'.format(b) for b in x)
    else:
        b = value.to_bytes(math.ceil(length/8), byteorder='big',
                             signed=signed)
        b = '{:0{}b}'.format(int.from_bytes(b, byteorder='big'),
                             length)
        bitstring = b[:length]

    return bitstring


@functools.lru_cache(4096)
def unpack_bitstring(length, is_float, is_signed, bits):
    if is_float:
        # TODO: CAMPid 097897541967932453154321546542175421549
        types = {
            32: '>f',
            64: '>d'
        }

        float_type = types.get(length)

        if float_type is None:
            raise Exception(
                'float type only supports lengths in [{}]'.
                    format(', '.join([str(t) for t in types.keys()]))
            )

        value, = struct.unpack(
            float_type,
            bytes(int(''.join(b), 2)
                      for b in epyqlib.utils.general.grouper(bits, 8))
        )
    else:
        value = int(bits, 2)

        if is_signed and bits[0] == '1':
            value -= (1 << len(bits))

    return value


@functools.lru_cache(4096)
def bytes_to_bitstrings(data):
    b = tuple('{:08b}'.format(b) for b in data)
    little = ''.join(reversed(b))
    big = ''.join(b)

    return little, big


@functools.lru_cache(4096)
def bitstring_to_signal_list(signals, big, little):
    unpacked = []
    for signal in signals:
        if signal.little_endian:
            least = 64 - signal.start_bit
            most = least - signal.signal_size

            bits = little[most:least]
        else:
            most = signal.start_bit
            least = most + signal.signal_size

            bits = big[most:least]

        unpacked.append(signal.unpack_bitstring(bits))
    return unpacked


@functools.lru_cache(4096)
def signals_to_bytes(length, signals, data):
    little_bits = [None] * (length * 8)
    big_bits = list(little_bits)
    for value, signal in zip(data, signals):
        bits = signal.pack_bitstring(value)

        if signal.little_endian:
            least = 64 - signal.start_bit
            most = least - signal.signal_size

            little_bits[most:least] = bits
        else:
            most = signal.start_bit
            least = most + signal.signal_size

            big_bits[most:least] = bits
    little_bits = reversed(tuple(epyqlib.utils.general.grouper(
        little_bits, 8)))
    little_bits = tuple(itertools.chain(*little_bits))
    bitstring = ''.join(
        next(x for x in (l, b, '0') if x is not None)
        # l if l != ' ' else (b if b != ' ' else '0')
        for l, b in zip(little_bits, big_bits)
    )
    return bytes(
        int(''.join(b), 2)
        for b in epyqlib.utils.general.grouper(bitstring, 8)
    )


class Signal:
    # TODO: but some (progress bar, etc) require an int!
    value_changed = epyqlib.utils.qt.Signal(float)
    value_set = epyqlib.utils.qt.Signal(float)

    def __init__(self, signal, frame, connect=None, parent=None):
        # self.attributes = signal._attributes # {dict} {'GenSigStartValue': '0.0', 'LongName': 'Enable'}
        self.default_value = signal.initial_value
        self.long_name = signal.attributes.get('LongName', None)
        self.hexadecimal_output = signal.attributes.get('HexadecimalOutput',
                                                         None)
        self.hexadecimal_output = self.hexadecimal_output is not None
        self.little_endian = signal.is_little_endian # {int} 0
        self.comment = signal.comment # {str} 'Run command.  When set to a value of \\'Enable\\', causes transition to grid forming or grid following mode depending on whether AC power is detected.  Must be set to \\'Disable\\' to leave POR or FAULTED state.'
        if self.comment is None:
            self.comment = ''
        # TODO: CAMPid 03549854754276996754265427 (repeated <summary> check)
        self.is_summary = '<summary>' in self.comment
        # TODO: maybe not use a string, but used to help with decimal places
        self.factor = signal.factor
        try:
            self.max = signal.max
        except ValueError:
            # TODO: default based on signal range
            self.max = None
        try:
            self.min = signal.min
        except ValueError:
            # TODO: default based on signal range
            self.min = None
        try:
            self.offset = signal.offset
        except ValueError:
            self.offset = 0

        if signal.multiplex == 'Multiplexor':
            self.multiplex = True
        else:
            self.multiplex = signal.multiplex # {NoneType} None

        self.raw_minimum, self.raw_maximum = signal.calculate_raw_range()

        self.name = signal.name # {str} 'Enable_command'
        # self.receiver = signal.receiver # {str} ''
        self.signal_size = int(signal.size) # {int} 2
        self.start_bit = int(signal.get_startbit()) # {int} 0
        self.unit = signal.unit # {str} ''
        self.enumeration = {int(k): v for k, v in signal.values.items()} # {dict} {'0': 'Disable', '2': 'Error', '1': 'Enable', '3': 'N/A'}
        self.enumeration_name = signal.enumeration
        self.signed = signal.is_signed
        if self.multiplex is True:
            self.signed = False
        self.float = signal.is_float

        self._format = None

        self.value = None
        self.scaled_value = None

        self.frame = frame
        if self.frame is not None:
            # TODO: put this into the frame!
            self.frame.signals.append(self)

        self.enumeration_format_re = {'re': r'^\[(\d+)\]',
                                      'format': '[{v}] {s}',
                                      'no_value_format': '{s}'}

        # TODO: make this configurable in the .sym?
        self.secret = self.name.casefold() in {'factoryaccess', 'password'}

        self.decimal_places = None

        self._format = None

        self.full_string, self.short_string, self.enumeration_text = (
            self.format_strings(value=self.value)
        )
        if connect is not None:
            self.connect(connect)

        if signal.comment is None:
            self.parameter_uuid = None
        else:
            self.comment, self.parameter_uuid = strip_uuid_from_comment(
                self.comment,
            )

    def __str__(self):
        return '{name}: sb:{start_bit}, osb:{ordering_start_bit}, len:{length}'.format(
            name=self.name,
            start_bit=self.start_bit,
            ordering_start_bit=getattr(self, 'ordering_start_bit', '-'),
            length=self.signal_size
        )

    def last_received(self):
        return self.frame.last_received

    def to_human(self, value):
        return self.offset + (value * self.factor)

    def from_human(self, value):
        return round((value - self.offset) / self.factor)

    def get_human_value(self, for_file=False, column=None, value=nothing):
        if value is nothing:
            value = self.value

        # TODO: handle offset
        if value is None:
            return None

        value = self.to_human(value)

        return self.format_float(value, for_file=for_file)

    def calc_human_value(self, raw_value):
        # TODO: handle offset
        locale.setlocale(locale.LC_ALL, '')

        if isinstance(raw_value, str):
            enumeration_strings = self.enumeration_strings()
            if len(enumeration_strings) > 0:
                try:
                    index = enumeration_strings.index(raw_value)
                except ValueError:
                    enumeration_strings = self.enumeration_strings(
                        include_values=True)
                    if len(enumeration_strings) > 0:
                        try:
                            index = enumeration_strings.index(raw_value)
                        except ValueError:
                            index = int(raw_value)

                value = index
            else:
                value = decimal.Decimal(locale.delocalize(raw_value))
        else:
            value = raw_value

        if value is not None:
            value = self.from_human(value)

        return value

    def check_human_value(self, raw_value, check_range=False):
        value = self.calc_human_value(raw_value)

        return self.check_value(
            value=value,
            check_range=check_range,
        )

    def set_human_value(self, raw_value, force=False, check_range=False):
            value = self.calc_human_value(raw_value)

            self.set_value(
                value=value,
                force=force,
                check_range=check_range,
            )

    def enumeration_string(self, value, include_value=False):
        format = (self.enumeration_format_re['format']
                  if include_value
                  else self.enumeration_format_re['no_value_format'])

        return format.format(v=value, s=self.enumeration[value])

    def enumeration_strings(self, include_values=False):
        items = list(self.enumeration)
        items.sort()
        items = [self.enumeration_string(i, include_value=include_values)
                 for i in items]

        return items

    def get_decimal_places(self):
        if self.decimal_places is None:
            if self.float:
                # TODO: these signals probably ought to have decimal places
                #       specified in the .sym, but that isn't supported yet
                #       anyways.
                self.decimal_places = 3
            else:
                x = self.factor
                # http://stackoverflow.com/a/3019027/228539
                max_digits = 14
                int_part = int(abs(x))
                magnitude = 1 if int_part == 0 else int(math.log10(int_part)) + 1
                if magnitude >= max_digits:
                    return (magnitude, 0)
                frac_part = abs(x) - int_part
                multiplier = 10 ** (max_digits - magnitude)
                frac_digits = multiplier + int(multiplier * frac_part + decimal.Decimal('0.5'))
                while frac_digits % 10 == 0:
                    frac_digits /= 10
                scale = int(math.log10(frac_digits))

                self.decimal_places = scale

        return self.decimal_places

    def check_value(self, value, check_range=False, minimum=None, maximum=None):
        if minimum is None:
            minimum = self.min

        if maximum is None:
            maximum = self.max

        if type(value) is float and math.isnan(value):
            return False

        if check_range:
            human_value = self.to_human(value)
            if not minimum <= human_value <= maximum:
                raise OutOfRangeError('{} not in range [{}, {}]'.format(
                    *[self.format_float(f) for f
                      in (human_value, minimum, maximum)]))

        return True

    def set_value(self, value, force=False, check_range=False, minimum=None,
                  maximum=None):
        ok = Signal.check_value(
            self=self,
            value=value,
            check_range=check_range,
            minimum=minimum,
            maximum=maximum,
        )
        if not ok:
            return False

        value_parameter = value

        if self.value != value or force:
            self.value = value

            if value not in self.enumeration:
                # TODO: this should be a subclass or something
                if self.hexadecimal_output:
                    pass
                elif self.value is not None:
                    # TODO: CAMPid 9395616283654658598648263423685
                    # TODO: and _offset...

                    self.scaled_value = (
                        self.offset + (self.value * self.factor)
                    )

                    value = self.scaled_value

            self.full_string, self.short_string, self.enumeration_text = (
                self.format_strings(value=value_parameter)
            )

            if value_parameter is None:
                self.value_changed.emit(float('nan'))
            else:
                self.value_changed.emit(value)

        if value is not None:
            self.value_set.emit(value)

    def format_strings(self, value):
        if value is None or (type(value) is float and math.isnan(value)):
            full_string = '-'
            short_string = '-'
            enumeration_text = None
        else:
            enumeration_text = None

            try:
                # TODO: CAMPid 94562754956589992752348667
                enum_string = self.enumeration[value]
                full_string = self.enumeration_format_re['format'].format(
                        s=enum_string, v=value)
                enumeration_text = enum_string
                short_string = enum_string
            except KeyError:
                # TODO: this should be a subclass or something
                if self.hexadecimal_output:
                    format = '{{:0{}X}}'.format(math.ceil(self.signal_size/math.log2(16)))
                    full_string = format.format(int(value))
                    short_string = full_string
                else:
                    # TODO: CAMPid 9395616283654658598648263423685
                    # TODO: and _offset...

                    scaled_value = (
                        self.offset + (value * self.factor)
                    )

                    full_string = self.format_float(scaled_value)
                    short_string = full_string

                    if self.unit is not None:
                        if len(self.unit) > 0:
                            full_string += ' [{}]'.format(self.unit)

        return full_string, short_string, enumeration_text

    def force_value_changed(self):
        value = self.scaled_value
        if value is None:
            value = 0
        self.value_changed.emit(value)

    def format_float(self, value=None, decimal_places=None, for_file=False):
        if self.secret:
            return '<secret>'

        # TODO: ack fix this since it's getting called with an actual None value...
        if value is None:
            value = self.scaled_value

        if value is None:
            formatted = '-'
        else:
            if decimal_places is None:
                decimal_places = self.get_decimal_places()

            if for_file:
                format = '{{:.{}f}}'.format(self.get_decimal_places())
                formatted = format.format(value)
            else:
                format = '%.{}f'.format(decimal_places)
                formatted = locale_format(format, value)

        return formatted

    def pack_bitstring(self, value=None):
        if value is None:
            value = self.value
        if value is None:
            value = 0

        try:
            return pack_bitstring(self.signal_size, self.float, value, self.signed)
        except OverflowError as e:
            names = (self.frame.name, self.frame.mux_name, self.name)
            name = ':'.join(name for name in names if name is not None)
            raise UnableToPackError(
                'Unable to pack {value} into {name} with range '
                '[{minimum}, {maximum}]'.format(
                    value=value,
                    name=name,
                    minimum=self.raw_minimum,
                    maximum=self.raw_maximum,
                )
            ) from e

    def unpack_bitstring(self, bits):
        return unpack_bitstring(self.signal_size, self.float, self.signed, bits)

    def some_packable_value(self):
        return self.unpack_bitstring(bits='0' * self.signal_size)


@functools.lru_cache(10000)
def locale_format(format, value):
    return locale.format_string(format, value, grouping=True)


class QtCanListener(can.Listener):
    message_received_signal = epyqlib.utils.qt.Signal(can.Message)

    def __init__(self, receiver=None, parent=None):
        can.Listener.__init__(self)

        if receiver is not None:
            self.receiver(receiver)

    def receiver(self, slot):
        self.message_received_signal.connect(slot)

    def on_message_received(self, msg):
        # TODO: Be careful since this is no longer being deep copied.
        #       It seems safe based on looking at the socketcan and
        #       pcan bus objects that construct a new Message() for
        #       each one received.  The Notifier loop just forgets
        #       about the message as soon as it is sent here.
        #
        #       This optimization is being justified by the 25% drop
        #       in CPU usage.

        self.message_received_signal.emit(msg)

    def move_to_thread(self, thread):
        signal = type(self).message_received_signal
        signal.qobject_host(self).moveToThread(thread)


class Frame(QtCanListener):
    send = epyqlib.utils.qt.Signal(can.Message, 'PyQt_PyObject')

    def __init__(self, frame, multiplex_value=None,
                 signal_class=Signal, set_value_to_default=True,
                 mux_frame=None, strip_summary=True, parent=None):
        super().__init__(self.message_received, parent=parent)

        self.mux_frame = mux_frame

        self.id = frame.arbitration_id.id # {int} 16755521
        # self.SignalGroups = frame.SignalGroups # {list} []
        self.size = frame.size # {int} 8
        # self.Transmitter = frame.Transmitter # {list} []
        # self.attributes = frame.attributes # {dict} {'GenMsgCycleTime': '200'}
        self.cycle_time = frame.cycle_time if frame.cycle_time != 0 else None
        self.mux_name = frame.attributes.get('mux_name', None)
        self.sendable = frame.attributes.get('Sendable') == 'True'
        self.receivable = frame.attributes.get('Receivable') == 'True'
        self.comment = frame.comment # {str} 'Operational commands are received by the module via control bits within this message.'
        if self.comment is None:
            self.comment = ''
        self.extended = bool(frame.arbitration_id.extended) # {int} 1
        self.name = frame.name # {str} 'CommandModeControl'
        # self.receiver = frame.receiver # {list} []
        # self.signals = frame.signals # {list} [<canmatrix.canmatrix.Signal object at 0x7fddf8053fd0>, <canmatrix.canmatrix.Signal object at 0x7fddf8054048>, <canmatrix.canmatrix.Signal object at 0x7fddf80543c8>, <canmatrix.canmatrix.Signal object at 0x7fddf8054470>, <canmatrix.canmatrix.Signal object

        self._cyclic_requests = {}
        self._cyclic_period = None
        self.user_send_control = True
        self.block_cyclic = False
        self.timer = QTimer()
        self.timer.timeout.connect(self._update_and_send)

        self.format_str = None
        self.data = None
        self.last_received = None

        self.signals = []
        for signal in frame.signals:
            # TODO: CAMPid 03549854754276996754265427 (repeated <summary> check)
            skipping_summary_signal = (
                strip_summary
                and signal.comment is not None
                and '<summary>' in signal.comment
            )
            if skipping_summary_signal:
                continue

            if (multiplex_value is None or
                        str(signal.multiplex) == multiplex_value):
                neo_signal = signal_class(signal=signal, frame=self)

                factor = neo_signal.factor
                if factor is None:
                    factor = 1

                offset = neo_signal.offset
                if offset is None:
                    offset = decimal.Decimal('0')

                default_value = neo_signal.default_value
                if default_value is None:
                    default_value = 0

                if set_value_to_default:
                    neo_signal.set_human_value(
                        offset + (default_value * factor))

        self.signals = tuple(self.signals)

        self.mux_value = None
        if self.mux_name is not None:
            for signal in self.signals:
                if not isinstance(signal.multiplex, bool):
                    self.mux_value = signal.multiplex
                    break

    def _update_and_send(self):
        if not self.block_cyclic:
            self._send(update=True)

    def signal_by_name(self, name):
        try:
            return next(s for s in self.signals if s.name == name)
        except StopIteration:
            return None

    def update_from_signals(self, function=None, data=None, only_return=False):
        if data is None:
            data = self

        data = self.pack(data, function=function)

        if not only_return:
            self.data = data

        return data

    def pack(self, data, function=None):
        if data == self:
            if function is None:
                function = lambda s: s.value
            data = []
            for signal in self.signals:
                try:
                    value = function(signal)
                except:
                    value = 0

                try:
                    value = int(value)
                except (TypeError, ValueError):
                    value = 0
                data.append(value)
            data = tuple(data)

        return signals_to_bytes(self.size, self.signals, data)

    def unpack(self, data, report_error=True, only_return=False):
        rx_length = len(data)
        if rx_length != self.size and report_error:
            print('Received message 0x{self.id:08X} with length {rx_length}, expected {self.size}'.format(**locals()))
        else:
            little, big = bytes_to_bitstrings(bytes(data))

            unpacked = bitstring_to_signal_list(self.signals, big, little)

            if only_return:
                return dict(zip(self.signals, unpacked))

            for s, v in zip(self.signals, unpacked):
                s.set_value(v)

    def _send(self, update=False):
        if update:
            self.data = self.pack(self)

        self.send.emit(self.to_message(), None)

    def _sent(self):
        pass

    def send_now(self, signals=None):
        self._send(update=True)

    def cyclic_request(self, caller, period):
        if period is None:
            try:
                del self._cyclic_requests[caller]
            except KeyError:
                pass
        else:
            # period will be able to converted to a float, test
            # sooner rather than later for easier debugging
            float(period)
            self._cyclic_requests[caller] = period

        periods = [float(v) for v in self._cyclic_requests.values()]
        new_period = min(periods) if len(periods) > 0 else None

        if new_period is not None:
            if new_period <= 0:
                new_period = None

        if new_period != self._cyclic_period:
            self._cyclic_period = new_period

            if self._cyclic_period is None:
                self.timer.stop()
            else:
                self.timer.setInterval(
                    int(round(float(self._cyclic_period) * 1000)))
                if not self.timer.isActive():
                    self.timer.start()

    def to_message(self, data=None):
        if data is None:
            data = self.data

        return can.Message(extended_id=self.extended,
                           arbitration_id=self.id,
                           dlc=self.size,
                           data=data)

    def message_received(self, msg):
        unpacked = False

        if (msg.arbitration_id == self.id and
                bool(msg.is_extended_id) == self.extended):
            if self.mux_frame is None:
                self.unpack(msg.data)
                unpacked = True
            elif self.mux_frame is self:
                # print(self, self.name, self.mux_name, self.mux_frame, self.mux_frame.name, self.mux_frame.mux_name)

                unpacked = self.mux_frame.unpack(msg.data, only_return=True)
                mux_signal, = unpacked

                # TODO: this if added to avoid exceptions temporarily
                if mux_signal.value in self.multiplex_frames:
                    unpacked = (
                        self.multiplex_frames[mux_signal.value]
                            .message_received(msg)
                    )
            else:
                self.unpack(msg.data)
                unpacked = True

        if unpacked:
            self.last_received = msg.timestamp

        return unpacked

    def terminate(self):
        callers = tuple(r for r in self._cyclic_requests)
        for caller in callers:
            self.cyclic_request(caller, None)

        logging.debug('{} terminated'.format(object.__repr__(self)))


@functools.lru_cache(1024)
def frame_by_id(id, frames):
    found = (
        f for f in frames
        if f.id == id and f.mux_name is None
    )

    try:
        frame, = found
    except ValueError:
        return None

    return frame


class Neo(QtCanListener):
    def __init__(self, matrix, frame_class=Frame, signal_class=Signal,
                 rx_interval=0, bus=None, node_id_adjust=None,
                 strip_summary=True, parent=None):
        super().__init__(self.message_received, parent=parent)

        self.bus = None

        self.frame_rx_timestamps = {}
        self.frame_rx_interval = rx_interval

        frames = []

        for frame in matrix.frames:
            if node_id_adjust is not None:
                frame.arbitration_id = canmatrix.canmatrix.ArbitrationId(
                    id=node_id_adjust(
                        message_id=frame.arbitration_id.id,
                        to_device=(
                            frame.attributes['Receivable'].casefold() == 'false'
                        ),
                    ),
                    extended=frame.arbitration_id.extended,
                )
            multiplex_signal = None
            for signal in frame.signals:
                if signal.multiplex == 'Multiplexor':
                    multiplex_signal = signal
                    break

            if multiplex_signal is None:
                neo_frame = frame_class(
                    frame=frame,
                    strip_summary=strip_summary,
                )
                # for signal in frame.signals:
                #     signal = signal_class(signal=signal, frame=neo_frame)
                #     signal.set_human_value(signal.default_value *
                #                            signal.factor[float])
                frames.append(neo_frame)
            else:
                # Make a frame with just the multiplexor entry for
                # parsing messages later
                multiplex_frame = canmatrix.Frame(
                    name=frame.name,
                    arbitration_id=frame.arbitration_id,
                    size=frame.size,
                    transmitters=list(frame.transmitters),
                    cycle_time=frame.cycle_time,
                )
                matrix_signal = canmatrix.Signal(
                        name=multiplex_signal.name,
                        start_bit=multiplex_signal.start_bit,
                        size=multiplex_signal.size,
                        is_little_endian=multiplex_signal.is_little_endian,
                        is_signed=multiplex_signal.is_signed,
                        factor=multiplex_signal.factor,
                        offset=multiplex_signal.offset,
                        min=multiplex_signal.min,
                        max=multiplex_signal.max,
                        unit=multiplex_signal.unit,
                        multiplex=multiplex_signal.multiplex)
                multiplex_frame.add_signal(matrix_signal)
                multiplex_neo_frame = frame_class(
                    frame=multiplex_frame,
                    strip_summary=strip_summary,
                )
                multiplex_neo_frame.mux_frame = multiplex_neo_frame
                frames.append(multiplex_neo_frame)

                multiplex_neo_frame.multiplex_signal =\
                    multiplex_neo_frame.signals[0]

                multiplex_neo_frame.multiplex_frames = {}

                for multiplex_value, multiplex_name in multiplex_signal.values.items():
                    # For each multiplexed frame, make a frame with
                    # just those signals.
                    matrix_frame = canmatrix.Frame(
                        name=frame.name,
                        arbitration_id=frame.arbitration_id,
                        size=frame.size,
                        transmitters=list(frame.transmitters),
                        cycle_time=frame.cycle_time,
                    )
                    matrix_frame.add_attribute('mux_name', multiplex_name)
                    matrix_frame.add_comment(multiplex_signal.comments[int(
                        multiplex_value)])
                    matrix_signal = canmatrix.Signal(
                            name=multiplex_signal.name,
                            start_bit=multiplex_signal.start_bit,
                            size=multiplex_signal.size,
                            is_little_endian=multiplex_signal.is_little_endian,
                            is_signed=multiplex_signal.is_signed,
                            factor=multiplex_signal.factor,
                            offset=multiplex_signal.offset,
                            min=multiplex_signal.min,
                            max=multiplex_signal.max,
                            unit=multiplex_signal.unit,
                            multiplex=multiplex_signal.multiplex)
                    # neo_signal = signal_class(signal=matrix_signal, frame=multiplex_neo_frame)
                    matrix_frame.add_signal(matrix_signal)

                    for signal in frame.signals:
                        if signal.multiplex == multiplex_value:
                            matrix_frame.add_signal(signal)

                    neo_frame = frame_class(
                        frame=matrix_frame,
                        mux_frame=multiplex_neo_frame,
                        strip_summary=strip_summary,
                    )
                    for signal in neo_frame.signals:
                        if signal.multiplex is True:
                            signal.set_value(multiplex_value)
                    frames.append(neo_frame)
                    multiplex_neo_frame.\
                        multiplex_frames[multiplex_value] = neo_frame

        self.frames = tuple(frames)

        self.signal_from_uuid = {
            signal.parameter_uuid: signal
            for frame in self.frames
            for signal in frame.signals
            if signal.parameter_uuid is not None
        }

        if bus is not None:
            self.set_bus(bus=bus)

    def set_bus(self, bus):
        if self.bus is not None:
            raise Exception('Bus already set')

        self.bus = bus

        for frame in self.frames:
            frame.send.connect(self.bus.send)

    def frame_by_id(self, id):
        return frame_by_id(id, self.frames)

    def frame_by_name(self, name):
        try:
            return next(f for f in self.frames if f.name == name)
        except StopIteration:
            return None

    def signal_by_path(self, *elements):
        i = iter(elements)

        def get_next(i):
            try:
                return next(i)
            except StopIteration as e:
                raise NotFoundError(repr(elements)) from e

        element = get_next(i)

        frame = self.frame_by_name(element)

        if frame is None:
            raise NotFoundError(repr(elements))

        if hasattr(frame, 'multiplex_frames'):
            element = get_next(i)

            frames = (
                f for f in frame.multiplex_frames.values()
                if f.mux_name == element
            )
            try:
                [frame] = frames
            except ValueError:
                raise NotFoundError(repr(elements))

        element = get_next(i)

        signal = frame.signal_by_name(element)
        if signal is None:
            raise NotFoundError(repr(elements))

        return signal

    def get_multiplex(self, message):
        base_frame = self.frame_by_id(message.arbitration_id)

        if not hasattr(base_frame, 'multiplex_frames'):
            frame = base_frame
            multiplex_value = None
        else:
            # finish the multiplex thing
            base_frame.unpack(message.data, report_error=False)
            multiplex_value = base_frame.multiplex_signal.value
            try:
                frame = base_frame.multiplex_frames[multiplex_value]
            except KeyError:
                return (base_frame, None)

        return (frame, multiplex_value)

    def message_received(self, msg):
        frame = self.frame_by_id(msg.arbitration_id)
        if frame is not None:
            last = self.frame_rx_timestamps.get(frame,
                                                -self.frame_rx_interval)
            if msg.timestamp - last >= self.frame_rx_interval:
                self.frame_rx_timestamps[frame] = msg.timestamp
                frame.message_received_signal.emit(msg)

    def terminate(self):
        for frame in self.frames:
            frame.terminate()

        cached_functions = (
            pack_bitstring,
            unpack_bitstring,
            bytes_to_bitstrings,
            bitstring_to_signal_list,
            signals_to_bytes,
        )

        for f in cached_functions:
            logging.debug('epyqlib.canneo.{}(): {}'.format(
                f.__name__,
                f.cache_info(),
            ))
        logging.debug('{} terminated'.format(object.__repr__(self)))


@functools.lru_cache(128)
def format_identifier(identifier, extended):
    f = '0x{{:0{}X}}'

    if extended:
        f = f.format(8)
    else:
        f = f.format(3)

    return f.format(identifier)

@functools.lru_cache(1024)
def format_data(data):
    return ' '.join(['{:02X}'.format(byte) for byte in data])
